import typing as t

from hrflow_connectors.v2.connectors.recruitee.warehouse import RecruiteeWarehouse
from hrflow_connectors.v2.core.common import Direction, Entity, Mode
from hrflow_connectors.v2.core.connector import Connector, ConnectorType, Flow


def get_profile_cv_url(attachments: t.List[t.Dict]):
    if not attachments:
        return None
    cv_url = next(
        (attachment for attachment in attachments if attachment.get("type") == "resume")
    )["public_url"]
    return cv_url


def get_profile_links(urls: t.List[t.Dict]) -> t.Tuple[t.List[str], t.List[str]]:
    profile_links = []
    social_links = []

    profile_links = [e["url"] for e in urls if e.get("type") == "from_resume"]
    social_links = [e["url"] for e in urls if e.get("type") != "from_resume"]

    return profile_links, social_links


def format_hrflow_profile(hrflow_profile: t.Dict) -> t.Dict:
    hrflow_profile_info = hrflow_profile["info"]
    profile_links, social_links = get_profile_links(hrflow_profile_info["urls"])
    recruitee_candidate = dict(
        name=hrflow_profile_info["full_name"],
        remote_cv_url=get_profile_cv_url(hrflow_profile["attachments"]),
        emails=[hrflow_profile_info["email"]],
        phones=[hrflow_profile_info["phone"]],
        social_links=social_links,
        links=profile_links,
        cover_letter="",
        sources=[hrflow_profile["source"]["name"]],
    )
    return recruitee_candidate


def format_hrflow_profile_for_update(hrflow_profile: t.Dict) -> t.Dict:
    recruitee_candidate = format_hrflow_profile(hrflow_profile)
    recruitee_candidate["id"] = hrflow_profile["reference"]
    del recruitee_candidate["sources"]
    return recruitee_candidate


def format_archive_in_recruitee(hrflow_profile: t.Dict) -> t.Dict:
    return dict(id=hrflow_profile["reference"])


def get_candidate_tags(recruitee_candidate: t.Dict) -> t.List[t.Dict]:
    tag_keys = [
        "source",
        "sources",
        "positive_ratings",
        "grouped_open_question_answers",
        "cover_letter",
        "last_message_at",
        "is_hired",
    ]
    tags = [{"name": key, "value": recruitee_candidate.get(key)} for key in tag_keys]

    tags += [
        {"name": field.get("name") or field["kind"], "value": field["values"]}
        for field in recruitee_candidate.get("fields", [])
        if field["kind"]
        not in {
            "education",
            "experience",
            "language_skill",
            "skill",
            "date_of_birth",
            "address",
        }
    ]
    tags += [{"name": "tags", "value": recruitee_candidate.get("tags", [])}]
    return tags


def extract_field_values(fields: t.List[t.Dict], kind: str) -> t.List:
    return next((field["values"] for field in fields if field["kind"] == kind), [])


def format_recruitee_candidate(recruitee_candidate: t.Dict) -> t.Dict:
    fields = recruitee_candidate.get("fields", [])

    educations = [
        {
            "school": edu["school"],
            "date_start": edu["start_date"],
            "date_end": edu["end_date"],
            "description": edu["description"],
            "title": edu["major"],
            "location": {"text": "", "lat": None, "lng": None},
        }
        for edu in extract_field_values(fields, "education")
    ]
    experiences = [
        {
            "company": exp["company"],
            "date_start": exp["start_date"],
            "date_end": exp["end_date"],
            "description": exp["description"],
            "title": exp["title"],
            "location": {"text": exp["location"], "lat": None, "lng": None},
        }
        for exp in extract_field_values(fields, "experience")
    ]
    languages = [
        {"name": lang["language_name"], "value": lang["level"]}
        for lang in extract_field_values(fields, "language_skill")
    ]
    skills = [
        {"name": skill["text"], "value": None}
        for skill in extract_field_values(fields, "skills")
    ]

    address = extract_field_values(fields, "address")
    date_of_birth = extract_field_values(fields, "date_of_birth")
    gender = extract_field_values(fields, "gender")

    # Inline name splitting logic
    full_name = recruitee_candidate.get("name", "")
    name_parts = full_name.split()
    first_name = " ".join(name_parts[:-1]) if len(name_parts) > 1 else full_name
    last_name = name_parts[-1] if len(name_parts) > 1 else ""

    emails = recruitee_candidate.get("emails", [])
    phones = recruitee_candidate.get("phones", [])
    urls = [
        {"url": url, "type": "from_resume"}
        for url in recruitee_candidate.get("social_links", [])
        + recruitee_candidate.get("links", [])
    ]
    profile = {
        "reference": str(recruitee_candidate.get("id")),
        "text": recruitee_candidate.get("description"),
        "info": {
            "full_name": full_name,
            "first_name": first_name,
            "last_name": last_name,
            "email": emails[0] if emails else None,
            "phone": phones[0] if phones else None,
            "date_birth": date_of_birth[0]["date"] if date_of_birth else None,
            "gender": gender[0]["value"] if gender else None,
            "picture": recruitee_candidate.get("photo_url"),
            "urls": urls,
            "location": {
                "text": address[0]["text"] if address else "",
                "lat": None,
                "lng": None,
            },
        },
        "created_at": recruitee_candidate.get("created_at"),
        "updated_at": recruitee_candidate.get("updated_at"),
        "educations": educations,
        "experiences": experiences,
        "languages": languages,
        "skills": skills,
        "resume": {
            "raw": recruitee_candidate.get("cv_file"),
        },
        "tags": get_candidate_tags(recruitee_candidate),
    }
    return profile


def format_archive_in_hrflow(recruitee_candidate: t.Dict) -> t.Dict:
    return dict(reference=str(recruitee_candidate["id"]))


def get_job_tags(recruitee_job: t.Dict) -> t.List[t.Dict]:
    tag_keys = [
        "category",
        "department",
        "options_cv",
        "options_cover_letter",
        "experience",
        "education",
        "employment_type",
        "remote_option",
        "number_of_openings",
        "candidates_count",
        "disqualified_candidates_count",
        "qualified_candidates_count",
        "hired_candidates_count",
        "email_confirmation_body",
        "status",
        "closed_at",
    ]
    tags = [{"name": key, "value": recruitee_job.get(key)} for key in tag_keys]
    tags += [{"name": "offer_tags", "value": recruitee_job.get("offer_tags", [])}]

    return tags


def get_ranges_float(recruitee_job: t.Dict) -> t.List[t.Dict]:
    t = lambda name, value_min, value_max, unit: dict(
        name=name, value_min=value_min, value_max=value_max, unit=unit
    )
    salary = recruitee_job.get("salary", {})
    ranges_float = [
        t(
            "working hours",
            recruitee_job.get("min_hours"),
            recruitee_job.get("max_hours"),
            "Hours per week",
        ),
        t(
            "salary per {}".format(salary.get("period")),
            salary.get("min"),
            salary.get("max"),
            salary.get("currency"),
        ),
    ]
    return ranges_float


def format_job(recruitee_job: t.Dict) -> t.Dict:
    sections = [
        dict(
            name="job_requirements",
            title="Job Requirements",
            description=recruitee_job["requirements"],
        ),
        dict(
            name="job_description",
            title="Job Description",
            description=recruitee_job["description"],
        ),
        dict(
            name="job_highlights",
            title="Job Highlights",
            description=recruitee_job["highlight_html"],
        ),
    ]
    job = dict(
        name=recruitee_job.get("title"),
        reference=str(recruitee_job.get("id")),
        created_at=recruitee_job.get("created_at"),
        updated_at=recruitee_job.get("updated_at"),
        location=dict(
            lat=None,
            lng=None,
            text=recruitee_job.get("location"),
            fields=dict(
                city=recruitee_job.get("city"),
                postcode=recruitee_job.get("postal_code"),
            ),
        ),
        url=recruitee_job.get("url"),
        summary=recruitee_job.get("description"),
        requirements=recruitee_job.get("requirements"),
        sections=sections,
        tags=get_job_tags(recruitee_job),
        ranges_float=get_ranges_float(recruitee_job),
    )
    return job


DESCRIPTION = (
    "Recruitee is an innovative and intuitive recruitment software that empowers"
    " businesses to streamline hiring processes and attract top talent. Recruitee"
    " provides a comprehensive solution that ensures a seamless and efficient"
    " recruitment experience."
)

Recruitee = Connector(
    name="Recruitee",
    type=ConnectorType.ATS,
    subtype="recruitee",
    description=DESCRIPTION,
    url="https://recruitee.com/",
    warehouse=RecruiteeWarehouse,
    flows=(
        Flow(
            Mode.create,
            Entity.profile,
            Direction.inbound,
            format=format_recruitee_candidate,
        ),
        Flow(
            Mode.update,
            Entity.profile,
            Direction.inbound,
            format=format_recruitee_candidate,
        ),
        Flow(
            Mode.archive,
            Entity.profile,
            Direction.inbound,
            format=format_archive_in_hrflow,
        ),
        Flow(
            Mode.create,
            Entity.profile,
            Direction.outbound,
            format=format_hrflow_profile,
        ),
        Flow(
            Mode.update,
            Entity.profile,
            Direction.outbound,
            format=format_hrflow_profile_for_update,
        ),
        Flow(
            Mode.archive,
            Entity.profile,
            Direction.outbound,
            format=format_archive_in_recruitee,
        ),
        Flow(Mode.create, Entity.job, Direction.inbound, format=format_job),
        Flow(Mode.update, Entity.job, Direction.inbound, format=format_job),
        Flow(
            Mode.archive, Entity.job, Direction.inbound, format=format_archive_in_hrflow
        ),
    ),
)



import typing as t
from collections import Counter
from enum import Enum

from msgspec import Struct
from pydantic import BaseModel


# This decoder add support for Counter[AnyEnum] which is a
# specialized dict used in RunResult
# See https://jcristharif.com/msgspec/extending.html#mapping-to-from-native-types
def msgspec_dec_hook(type: type, obj: t.Any) -> t.Any:
    if (
        t.get_origin(type) is Counter
        and len(t.get_args(type)) == 1
        and issubclass((EnumModel := t.get_args(type)[0]), Enum)
    ):
        return Counter({EnumModel(key): value for key, value in obj.items()})
    else:
        raise NotImplementedError(
            f"Objects of type {type} are not supported"
        )  # pragma: nocover


class StoreNotInitializedError(Exception):
    pass


InternalStateT = t.TypeVar("InternalStateT", contravariant=True)


class SaveP(t.Protocol[InternalStateT]):
    def __call__(
        self, state: InternalStateT, key: str, data: t.Union[BaseModel, Struct]
    ) -> None:
        ...  # pragma: nocover


class LoadP(t.Protocol[InternalStateT]):
    def __call__(
        self,
        state: InternalStateT,
        key: str,
        parse_as: t.Union[type[BaseModel], type[Struct]],
    ) -> t.Union[BaseModel, Struct, None]:
        ...  # pragma: nocover


class BackendStore(Struct, t.Generic[InternalStateT]):
    name: str
    get_state: t.Callable[[], InternalStateT]
    saver: SaveP[InternalStateT]
    loader: LoadP[InternalStateT]
    state: t.Optional[InternalStateT] = None

    def init(self):
        self.state = self.get_state()

    def save(self, key: str, data: t.Union[BaseModel, Struct]):
        if self.state is None:
            raise StoreNotInitializedError("Backend not initialized: call store.init()")
        return self.saver(self.state, key, data)

    ParseAsT = t.TypeVar("ParseAsT", bound=t.Union[BaseModel, Struct])

    @t.overload
    def load(self, key: str, parse_as: type[ParseAsT]) -> t.Optional[ParseAsT]:
        ...  # pragma: nocover

    @t.overload
    def load(
        self,
        key: str,
        parse_as: t.Union[type[BaseModel], type[Struct]],
    ) -> t.Union[BaseModel, Struct, None]:
        ...  # pragma: nocover

    def load(
        self,
        key: str,
        parse_as: t.Union[type[BaseModel], type[Struct]],
    ):
        if self.state is None:
            raise StoreNotInitializedError("Backend not initialized: call store.init()")
        return self.loader(self.state, key, parse_as)



from typing import Any, List, Optional

from msgspec import Struct, field


class TeamtailorJobAttributes(Struct):
    title: str
    pitch: Optional[str]
    body: str
    created_at: str = field(name="created-at")
    updated_at: Optional[str] = field(name="updated-at")
    status: Optional[str]
    human_status: Optional[str] = field(name="human-status")
    language_code: Optional[str] = field(name="language-code")
    picture: Optional[str]
    external_application_url: Optional[str] = field(name="external-application-url")
    tags: List[str]
    remote_status: Optional[str] = field(name="remote-status")
    currency: Optional[str]
    salary_time_unit: Optional[str] = field(name="salary-time-unit")
    min_salary: Optional[int] = field(name="min-salary")
    max_salary: Optional[int] = field(name="max-salary")
    employment_type: Optional[str] = field(name="employment-type")
    employment_level: Optional[str] = field(name="employment-level")
    sharing_image_layout: Optional[str] = field(name="sharing-image-layout")
    template_name: Optional[str] = field(name="template-name")
    name_requirement: Optional[str] = field(name="name-requirement")
    resume_requirement: Optional[str] = field(name="resume-requirement")
    additional_files_requirement: Optional[str] = field(
        name="additional-files-requirement"
    )
    cover_letter_requirement: Optional[str] = field(name="cover-letter-requirement")
    phone_requirement: Optional[str] = field(name="phone-requirement")
    internal: Optional[bool]
    internal_name: Optional[str] = field(name="internal-name")
    pinned: Optional[bool]
    start_date: Optional[str] = field(name="start-date")
    end_date: Optional[str] = field(name="end-date")


class TeamtailorJob(Struct):
    attributes: TeamtailorJobAttributes
    links: Optional[dict]
    id: Any
    relationships: Optional[dict]
    type: str


class TeamtailorCandidateAttributes(Struct):
    first_name: str = field(name="first-name")
    last_name: str = field(name="last-name")
    created_at: str = field(name="created-at")
    updated_at: Optional[str] = field(name="updated-at")
    email: str
    phone: Optional[str]
    picture: Optional[str]
    linkedin_url: Optional[str] = field(name="linkedin-url")
    linkedin_uid: Optional[str] = field(name="linkedin-uid")
    linkedin_profile: Optional[str] = field(name="linkedin-profile")
    facebook_profile: Optional[str] = field(name="facebook-profile")
    facebook_id: Optional[str] = field(name="facebook-id")
    pitch: Optional[str]
    resume: Optional[str]
    original_resume: Optional[str] = field(name="original-resume")
    sourced: Optional[bool]
    connected: Optional[bool]
    internal: Optional[bool]
    referred: Optional[bool]
    referring_url: Optional[str] = field(name="referring-url")
    referring_site: Optional[str] = field(name="referring-site")
    unsubscribed: Optional[bool]
    tags: Optional[List[str]]


class TeamtailorCandidate(Struct):
    attributes: TeamtailorCandidateAttributes
    links: Optional[dict]
    id: str
    relationships: Optional[dict]
    type: str



import json
import typing as t
from dataclasses import dataclass, field
from enum import Enum
from logging import getLogger
from pathlib import Path

from hrflow_connectors.v2.core.common import Direction, Entity, Mode, Schema
from hrflow_connectors.v2.core.hrflow import HrFlowWarehouse
from hrflow_connectors.v2.core.msgspec_pydantic_compat import json_schema
from hrflow_connectors.v2.core.run import (
    ActionInitError,
    CallbackT,
    FormatT,
    LogicsT,
    Metadata,
    RunResult,
    run,
)
from hrflow_connectors.v2.core.templating import WORKFLOW, workflow
from hrflow_connectors.v2.core.utils import CONNECTORS_DIRECTORY, compute_logo_path
from hrflow_connectors.v2.core.warehouse import Aisle, Warehouse

default_logger = getLogger(__name__)


def is_lambda(fn: t.Callable):
    return fn.__name__ == (lambda: None).__name__


EventParserT = t.Callable[[dict], dict]


class NoLambdaEventParser(Exception):
    pass


@dataclass
class Flow:
    mode: Mode
    entity: Entity
    direction: Direction
    override_name: t.Optional[str] = None
    format: t.Optional[FormatT] = None
    logics: t.Optional[LogicsT] = None
    callback: t.Optional[CallbackT] = None
    event_parser: t.Optional[EventParserT] = None

    def __post_init__(self):
        if self.event_parser is not None and is_lambda(self.event_parser):
            raise NoLambdaEventParser(
                "event_parser if supplied should not be a lambda "
                "function: Please use a regular 'def' function"
            )

    def default_name(self, connector_subtype: str):
        return (
            f"{self.mode.value}_{self.entity.name}s"
            f"_in_{'hrflow' if self.direction is Direction.inbound else connector_subtype}"  # noqa E501
        )

    def name(self, connector_subtype: str):
        return self.override_name or self.default_name(connector_subtype)

    def description(self, connector_name: str) -> str:
        mode: Mode = self.mode
        if mode is Mode.create:
            state = "created"
        elif mode is Mode.update:
            state = "updated"
        elif mode is Mode.archive:
            state = "archived"

        if self.direction is Direction.inbound:
            origin = connector_name
            target = "HrFlow"
        else:
            origin = "HrFlow"
            target = connector_name

        return (
            f"Send **{state}** '{self.entity.value}(s)' _from_ {origin} _to_ {target}"
        )


class ConnectorType(Enum):
    ATS = "ATS"
    CRM = "CRM"
    HCM = "HCM"
    Automation = "Automation"
    JobBoard = "Job Board"
    Classifieds = "Classified Ads"
    Other = "Other"


class InvalidFlow(Exception):
    pass


class MetadataVariableKW(t.TypedDict):
    origin_name: str
    target_name: str


class RunVariableKW(t.TypedDict):
    origin: Aisle
    origin_auth: dict
    origin_auth_schema: Schema
    target: Aisle
    target_auth: dict
    target_auth_schema: Schema


class PublicActionInterface(t.Protocol):
    def __call__(
        self,
        *,
        workflow_id: str,
        connector_auth: dict,
        hrflow_auth: dict,
        pull_parameters: dict,
        push_parameters: dict,
        init_error: t.Optional[ActionInitError] = None,
        format: t.Optional[FormatT] = None,
        logics: t.Optional[LogicsT] = None,
        callback: t.Optional[CallbackT] = None,
        persist: bool = True,
        incremental: bool = False,
    ) -> RunResult:
        ...  # pragma: nocover


def make_action(
    *,
    flow: Flow,
    connector_name: str,
    connector_subtype: str,
    hrflow_aisle: Aisle,
    connector_aisle: Aisle,
    connector_auth_schema: Schema,
) -> PublicActionInterface:
    LocalHrflowWarehouse = HrFlowWarehouse
    if flow.direction is Direction.inbound:
        metadata_kw = MetadataVariableKW(
            origin_name=connector_name,
            target_name="hrflow",
        )
    else:
        metadata_kw = MetadataVariableKW(
            origin_name="hrflow",
            target_name=connector_name,
        )

    def action(
        *,
        workflow_id: str,
        connector_auth: dict,
        hrflow_auth: dict,
        pull_parameters: dict,
        push_parameters: dict,
        init_error: t.Optional[ActionInitError] = None,
        format: t.Optional[FormatT] = None,
        logics: t.Optional[LogicsT] = None,
        callback: t.Optional[CallbackT] = None,
        persist: bool = True,
        incremental: bool = False,
    ):
        if flow.direction is Direction.inbound:
            run_kw = RunVariableKW(
                origin=connector_aisle,
                origin_auth=connector_auth,
                origin_auth_schema=connector_auth_schema,
                target=hrflow_aisle,
                target_auth=hrflow_auth,
                target_auth_schema=LocalHrflowWarehouse.auth,
            )
        else:
            run_kw = RunVariableKW(
                origin=hrflow_aisle,
                origin_auth=hrflow_auth,
                origin_auth_schema=LocalHrflowWarehouse.auth,
                target=connector_aisle,
                target_auth=connector_auth,
                target_auth_schema=connector_auth_schema,
            )

        return run(
            workflow_id=workflow_id,
            metadata=Metadata(
                connector_name=connector_name,
                action_name=flow.name(connector_subtype),
                using_default_format=format is None,
                using_default_logics=logics is None,
                **metadata_kw,
            ),
            mode=flow.mode,
            origin_parameters=pull_parameters,
            target_parameters=push_parameters,
            incremental=incremental,
            init_error=init_error,
            format=format if format is not None else flow.format,
            logics=logics if logics is not None else flow.logics,
            callback=callback if callback is not None else flow.callback,
            persist=persist,
            **run_kw,
        )

    return action


class WorkflowManifest(t.TypedDict):
    catch_template: str
    pull_template: str
    settings_keys: dict[str, str]
    placeholders: dict[str, str]
    expected: dict[str, str]


class ActionManifest(t.TypedDict):
    name: str
    data_type: str
    direction: t.Literal["inbound", "outbound"]
    mode: t.Literal["create", "update", "archive"]
    connector_auth_parameters: dict
    hrflow_auth_parameters: dict
    origin: str
    origin_data_schema: dict
    supports_incremental: bool
    pull_parameters: dict
    target: str
    target_data_schema: dict
    push_parameters: dict
    jsonmap: dict
    workflow: WorkflowManifest


class Manifest(t.TypedDict):
    name: str
    type: str
    subtype: str
    logo: str
    actions: list[ActionManifest]


@dataclass
class Connector:
    name: str
    subtype: str
    description: str
    url: str
    type: ConnectorType
    warehouse: Warehouse
    flows: tuple[Flow, ...] = field(default_factory=tuple)

    def __post_init__(self):
        for flow in self.flows:
            connector_aisle = self.warehouse.get_aisle(flow.entity)
            if connector_aisle is None:
                raise InvalidFlow(
                    f"Invalid flow {flow}: Entity={flow.entity} not supported by"
                    f" {self.name} warehouse"
                )

            hrflow_aisle = HrFlowWarehouse.get_aisle(flow.entity)
            if hrflow_aisle is None:
                raise InvalidFlow(
                    f"Invalid flow {flow}: Entity={flow.entity} not supported by HrFlow"
                    " warehouse"
                )

            if flow.direction is Direction.inbound:
                if connector_aisle.parameters("read", flow.mode) is None:
                    raise InvalidFlow(
                        f"Invalid flow {flow}: {self.name} warehouse is not readable in"
                        f" mode={flow.mode} for Entity={flow.entity}"
                    )
                if hrflow_aisle.parameters("write", flow.mode) is None:
                    raise InvalidFlow(
                        f"Invalid flow {flow}: HrFlow warehouse is not writable in"
                        f" mode={flow.mode} for Entity={flow.entity}"
                    )
            else:
                if hrflow_aisle.parameters("read", flow.mode) is None:
                    raise InvalidFlow(
                        f"Invalid flow {flow}: HrFlow warehouse is not readable in"
                        f" mode={flow.mode} for Entity={flow.entity}"
                    )
                if connector_aisle.parameters("write", flow.mode) is None:
                    raise InvalidFlow(
                        f"Invalid flow {flow}: {self.name} warehouse is not writable in"
                        f" mode={flow.mode} for Entity={flow.entity}"
                    )

            setattr(
                self,
                flow.name(self.subtype),
                make_action(
                    flow=flow,
                    connector_name=self.name,
                    connector_subtype=self.subtype,
                    hrflow_aisle=hrflow_aisle,
                    connector_aisle=connector_aisle,
                    connector_auth_schema=self.warehouse.auth,
                ),
            )

    def manifest(self, connectors_directory: Path = CONNECTORS_DIRECTORY) -> Manifest:
        actions: list[ActionManifest] = []
        manifest = Manifest(
            name=self.name,
            type=self.type.value.upper().replace(" ", ""),
            subtype=self.subtype,
            logo=compute_logo_path(
                name=self.name,
                subtype=self.subtype,
                connectors_directory=connectors_directory,
            ),
            actions=actions,
        )

        for flow in self.flows:
            hrflow_aisle = HrFlowWarehouse.get_aisle(flow.entity)
            connector_aisle = self.warehouse.get_aisle(flow.entity)

            # This is already validated in Connector.__post_init__
            assert hrflow_aisle is not None
            assert connector_aisle is not None

            jsonmap_path = (
                connectors_directory
                / self.subtype
                / "mappings"
                / "format"
                / "{}.json".format(flow.name(self.subtype))
            )
            try:
                jsonmap = json.loads(jsonmap_path.read_text())
            except FileNotFoundError:
                jsonmap = {}

            if flow.direction is Direction.inbound:
                origin_aisle = connector_aisle
                target_aisle = hrflow_aisle
            else:
                origin_aisle = hrflow_aisle
                target_aisle = connector_aisle

            pull_parameters = origin_aisle.parameters("read", flow.mode)
            push_parameters = target_aisle.parameters("write", flow.mode)

            # This is already validated in Connector.__post_init__
            assert origin_aisle.read is not None
            assert pull_parameters is not None
            assert push_parameters is not None

            action_manifest = ActionManifest(
                name=flow.name(self.subtype),
                data_type=flow.entity.value,
                direction=flow.direction.value,
                mode=flow.mode.value,
                connector_auth_parameters=json_schema(self.warehouse.auth),
                hrflow_auth_parameters=json_schema(HrFlowWarehouse.auth),
                origin=self.name if flow.direction is Direction.inbound else "HrFlow",
                origin_data_schema=json_schema(origin_aisle.schema),
                supports_incremental=origin_aisle.read.supports_incremental(flow.mode),
                pull_parameters=json_schema(pull_parameters),
                target="HrFlow" if flow.direction is Direction.inbound else self.name,
                target_data_schema=json_schema(target_aisle.schema),
                push_parameters=json_schema(push_parameters),
                jsonmap=jsonmap,
                workflow=WorkflowManifest(
                    catch_template=workflow(
                        connector=self, flow=flow, integration="catch"
                    ),
                    pull_template=workflow(
                        connector=self, flow=flow, integration="pull"
                    ),
                    settings_keys=dict(
                        workflow_id=WORKFLOW.WORKFLOW_ID_SETTINGS_KEY,
                        incremental=WORKFLOW.INCREMENTAL_SETTINGS_KEY,
                        connector_auth_prefix=WORKFLOW.CONNECTOR_AUTH_SETTINGS_PREFIX,
                        hrflow_auth_prefix=WORKFLOW.HRFLOW_AUTH_SETTINGS_PREFIX,
                        pull_parameters_prefix=WORKFLOW.PULL_PARAMETERS_SETTINGS_PREFIX,
                        push_parameters_prefix=WORKFLOW.PUSH_PARAMETERS_SETTINGS_PREFIX,
                    ),
                    placeholders=dict(
                        logics=WORKFLOW.LOGICS_PLACEHOLDER,
                        format=WORKFLOW.FORMAT_PLACEHOLDER,
                        callback=WORKFLOW.CALLBACK_PLACEHOLDER,
                        event_parser=WORKFLOW.EVENT_PARSER_PLACEHOLDER,
                    ),
                    expected=dict(
                        activate_incremental=WORKFLOW.ACTIVATE_INCREMENTAL,
                        logics_functions_name=WORKFLOW.LOGICS_FUNCTIONS_NAME,
                        format_functions_name=WORKFLOW.FORMAT_FUNCTION_NAME,
                        callback_functions_name=WORKFLOW.CALLBACK_FUNCTION_NAME,
                        event_parser_function_name=WORKFLOW.USER_EVENT_PARSER_FUNCTION_NAME,  # noqa E501
                    ),
                ),
            )
            actions.append(action_manifest)

        return manifest


def hrflow_connectors_manifest(
    connectors: t.Iterable[Connector],
    directory_path: str = ".",
    connectors_directory: Path = CONNECTORS_DIRECTORY,
) -> None:
    manifest = dict(
        name="HrFlow.ai Connectors",
        connectors=[
            connector.manifest(connectors_directory=connectors_directory)
            for connector in connectors
        ],
    )

    with open("{}/manifest.json".format(directory_path), "w") as f:
        f.write(json.dumps(manifest, indent=2))



import logging
import os
import secrets
import typing as t
from io import BytesIO

import boto3
import boto3.session
from msgspec import Struct, convert, defstruct, json
from pydantic import BaseModel, create_model

from hrflow_connectors.core.backend.common import BackendStore, msgspec_dec_hook

logger = logging.getLogger(__name__)

BUCKET_ENVIRONMENT_VARIABLE: t.Final = "HRFLOW_CONNECTORS_S3_BUCKET"
PREFIX_ENVIRONMENT_VARIABLE: t.Final = "HRFLOW_CONNECTORS_S3_PREFIX"
AWS_REGION_ENVIRONMENT_VARIABLE: t.Final = "HRFLOW_CONNECTORS_S3_AWS_REGION"
AWS_ACCESS_KEY_ID_ENVIRONMENT_VARIABLE: t.Final = (
    "HRFLOW_CONNECTORS_S3_AWS_ACCESS_KEY_ID"
)
AWS_SECRET_ACCESS_KEY_ENVIRONMENT_VARIABLE: t.Final = (
    "HRFLOW_CONNECTORS_S3_AWS_SECRET_ACCESS_KEY"
)
S3_FOLDER_NAME: t.Final = "hrflow_connectors/store"
NAME: t.Final = "s3"


class InternalStateBase(Struct):
    bucket: str
    key_prefix: str

    def s3_key(self, key: str) -> str:
        return "{}/{}.json".format(self.key_prefix.strip("/"), key)


if t.TYPE_CHECKING:
    from mypy_boto3_s3 import S3Client  # pragma: nocover

    class InternalState(InternalStateBase):  # pragma: nocover
        client: S3Client

else:

    class InternalState(InternalStateBase):
        client: t.Any


class CannotWriteToS3Error(Exception):
    pass


class CannotReadFromS3Error(Exception):
    pass


def check_store_pydantic(state: InternalState):
    root_model = create_model("S3StoreRoot", root="HrFlow Connectors", store=NAME)
    root_key = f"__root_{secrets.token_hex(4)}"
    root = root_model()
    try:
        save(state, root_key, root)
    except Exception as e:
        raise CannotWriteToS3Error(
            "Failed to check writing to S3 with error={}".format(e)
        )

    try:
        loaded = load(state, key=root_key, parse_as=root_model)
        assert loaded == root, "Loaded data should match original data pydantic"
    except Exception as e:
        raise CannotReadFromS3Error(
            "Failed to check reading from S3 with error={}".format(e)
        )


def check_store_msgspec(state: InternalState):
    root_model = defstruct(
        "S3StoreRoot", [("root", str, "HrFlow Connectors"), ("store", str, NAME)]
    )
    root_key = f"__root_{secrets.token_hex(4)}"
    root = root_model()
    try:
        save(state, root_key, root)
    except Exception as e:
        raise CannotWriteToS3Error(
            "Failed to check writing to S3 with error={}".format(e)
        )

    try:
        loaded = load(state, key=root_key, parse_as=root_model)
        assert loaded == root, "Loaded data should match original data msgspec"
    except Exception as e:
        raise CannotReadFromS3Error(
            "Failed to check reading from S3 with error={}".format(e)
        )


def check_store(state: InternalState):
    check_store_pydantic(state)
    check_store_msgspec(state)
    logger.info("S3 Backend properly configured")


def get_state():
    if (region_name := os.environ.get(AWS_REGION_ENVIRONMENT_VARIABLE)) is None:
        raise Exception(
            f"Missing environment variable {AWS_REGION_ENVIRONMENT_VARIABLE} in order"
            " to setup S3 store"
        )

    if (bucket := os.environ.get(BUCKET_ENVIRONMENT_VARIABLE)) is None:
        raise Exception(
            f"Missing environment variable {BUCKET_ENVIRONMENT_VARIABLE} for S3 store"
            " setup"
        )

    if os.environ.get(AWS_ACCESS_KEY_ID_ENVIRONMENT_VARIABLE) and os.environ.get(
        AWS_SECRET_ACCESS_KEY_ENVIRONMENT_VARIABLE
    ):
        logger.info("Configuring S3 Backend with explicit credentials")
        client = boto3.client(
            "s3",
            region_name=region_name,
            aws_access_key_id=os.environ.get(AWS_ACCESS_KEY_ID_ENVIRONMENT_VARIABLE),
            aws_secret_access_key=os.environ.get(
                AWS_SECRET_ACCESS_KEY_ENVIRONMENT_VARIABLE
            ),
        )
    else:
        logger.info("Configuring S3 Backend with implicit credentials")
        client = boto3.client(
            "s3",
            region_name=region_name,
        )

    key_prefix = S3_FOLDER_NAME
    prefix = os.environ.get(PREFIX_ENVIRONMENT_VARIABLE)
    if prefix is not None:
        key_prefix = "{}/{}".format(prefix.strip("/"), S3_FOLDER_NAME.strip("/"))

    state = InternalState(client=client, bucket=bucket, key_prefix=key_prefix)

    check_store(state)

    return state


def save(state: InternalState, key: str, data: t.Union[BaseModel, Struct]) -> None:
    if isinstance(data, BaseModel):
        with BytesIO(initial_bytes=data.json().encode()) as raw:
            state.client.upload_fileobj(raw, state.bucket, state.s3_key(key))
    else:
        with BytesIO(initial_bytes=json.encode(data)) as raw:
            state.client.upload_fileobj(raw, state.bucket, state.s3_key(key))


def load(
    state: InternalState, key: str, parse_as: t.Union[type[BaseModel], type[Struct]]
):
    with BytesIO() as raw:
        try:
            state.client.download_fileobj(state.bucket, state.s3_key(key), raw)
        except state.client.exceptions.ClientError as e:
            if int(e.response["Error"]["Code"]) == 404:
                return None
            raise e
        if issubclass(parse_as, BaseModel):
            return parse_as.parse_raw(raw.getvalue().decode())
        return convert(json.decode(raw.getvalue()), parse_as, dec_hook=msgspec_dec_hook)


S3Store = BackendStore(name=NAME, get_state=get_state, saver=save, loader=load)



from hrflow_connectors.v2.core.hrflow.aisles.common import (  # noqa: F401
    AuthParameters as AuthParameters,
)
from hrflow_connectors.v2.core.hrflow.aisles.job import (  # noqa: F401
    JobsAisle as JobsAisle,
)
from hrflow_connectors.v2.core.hrflow.aisles.profile import (  # noqa: F401
    ProfilesAisle as ProfilesAisle,
)



# This file is generated automatically
from hrflow_connectors.v2.core.connector import Connector, PublicActionInterface

class WaalaxyProto(Connector):
    create_profiles_in_hrflow: PublicActionInterface
    update_profiles_in_hrflow: PublicActionInterface
    archive_profiles_in_hrflow: PublicActionInterface
    
Waalaxy: WaalaxyProto


import typing as t

from hrflow_connectors.v2.connectors.breezyhr.utils import (
    is_valid_url,
    remove_html_tags,
)
from hrflow_connectors.v2.connectors.breezyhr.warehouse import BreezyHrWarehouse
from hrflow_connectors.v2.core.common import Direction, Entity, Mode
from hrflow_connectors.v2.core.connector import Connector, ConnectorType, Flow


def get_location(location: t.Dict) -> t.Dict:
    country = location.get("country", {}).get("name")
    city = location.get("city")
    address = location.get("name")
    street_address = location.get("streetAddress", {}).get("custom")
    geojson = dict(country=country, city=city, text=street_address or address)

    return dict(text=street_address or address, geojson=geojson, lat=None, lng=None)


def format_job(breezy_job: t.Dict) -> t.Dict:
    """
    Format a Breezy Hr job object into a hrflow job object
    Returns:
        HrflowJob: a job object in the hrflow job format
    """
    location = get_location(breezy_job.get("location", {}))
    is_remote = breezy_job.get("location", {}).get("is_remote")
    remote_details = breezy_job.get("location", {}).get("remote_details")
    description = remove_html_tags(breezy_job.get("description", ""))
    cleaned_description = description.replace("&nbsp;", " ")

    sections = [
        dict(
            name="description",
            title="Description",
            description=cleaned_description,
        ),
        dict(
            name="experience",
            title="Required Experience",
            description=breezy_job.get("experience", {}).get("name"),
        ),
        dict(
            name="education",
            title="Required Education",
            description=breezy_job.get("education"),
        ),
        dict(
            name="category",
            title="Category",
            description=breezy_job.get("category", {}).get("name"),
        ),
        dict(
            name="remote",
            title="Remote Information",
            description=remote_details,
        ),
    ]
    t = lambda name, value: dict(name=name, value=value)
    tags = [
        t("type", breezy_job.get("type", {}).get("name")),
        t("experience", breezy_job.get("experience", {}).get("name")),
        t("education", breezy_job.get("education")),
        t("department", breezy_job.get("department")),
        t("requisition_id", breezy_job.get("requisition_id")),
        t("category", breezy_job.get("category", {}).get("name")),
        t("candidate_type", breezy_job.get("candidate_type")),
        t("isremote", is_remote),
        t("remote_details", remote_details),
        t("creator_id", breezy_job.get("creator_id")),
        t("breezy_hr_tags", breezy_job.get("tags")),
    ]

    hrflow_job = dict(
        name=breezy_job.get("name"),
        reference=breezy_job.get("_id"),
        summary=cleaned_description,
        location=location,
        sections=sections,
        tags=tags,
        created_at=breezy_job.get("creation_date"),
        updated_at=breezy_job.get("updated_date"),
    )

    return hrflow_job


def format_archive_in_hrflow(breezy_element: t.Dict) -> t.Dict:
    return dict(reference=breezy_element.get("_id"))


def format_profile(hrflow_profile: t.Dict) -> t.Dict:
    """
    Format a Hrflow profile object into a breezy hr profile object
    Args:
        data (HrflowProfile): Hrflow Profile to format
    Returns:
        BreezyProfileModel: a BreezyHr formatted profile object
    """

    info = hrflow_profile.get("info", {})

    work_history = []
    for experience in hrflow_profile.get("experiences", []):
        formatted_experience = dict()
        if experience.get("company") not in ["", None]:
            formatted_experience["company_name"] = experience.get("company")
        else:
            formatted_experience["company_name"] = "Undefined"
        formatted_experience["title"] = experience.get("title")
        formatted_experience["summary"] = experience.get("description")
        if experience.get("date_start") is not None:
            date_start = experience["date_start"]
            formatted_experience["start_year"] = int(date_start[:4])
            formatted_experience["start_month"] = int(date_start[5:7])
        if experience.get("date_end") is not None:
            date_end = experience["date_end"]
            formatted_experience["end_year"] = int(date_end[:4])
            formatted_experience["end_month"] = int(date_end[5:7])
        work_history.append(formatted_experience)

    educations = []
    for education in hrflow_profile.get("educations", []):
        formatted_education = dict()
        if education.get("school") == "":
            education["school"] = "Undefined"
        formatted_education["school_name"] = education.get("school")
        formatted_education["field_of_study"] = education.get("title")
        if education.get("date_start") is not None:
            date_start = education["date_start"]
            formatted_education["start_year"] = int(date_start[:4])
        if education.get("date_end") is not None:
            date_end = education["date_end"]
            formatted_education["end_year"] = int(date_end[:4])
        educations.append(formatted_education)

    social_profiles = {}
    if info.get("urls"):
        for url in info.get("urls"):
            type = url.get("type")
            link = url.get("url")
            if type and link:
                if type == "from_resume" or not is_valid_url(link):
                    continue
                social_profiles.update({type: link})

    # add profile skills to tags
    tags = []
    skills = hrflow_profile.get("skills")
    if skills:
        tags = [skill.get("name") for skill in skills]

    # add resume to profile
    attachments = hrflow_profile.get("attachments", [])
    resume_url = next(
        (
            attachment
            for attachment in attachments
            if attachment.get("type") == "resume"
        ),
        {},
    ).get("public_url")
    breezy_profile = dict(
        name=info.get("full_name"),
        email_address=info.get("email"),
        phone_number=info.get("phone"),
        address=info.get("location", {}).get("text"),
        summary=info.get("summary"),
        work_history=work_history,
        education=educations,
        social_profiles=social_profiles,
        tags=tags,
        resume=resume_url,
    )
    return breezy_profile


def format_profile_for_update(hrflow_profile: t.Dict) -> t.Dict:
    breezy_profile = format_profile(hrflow_profile)
    breezy_profile["id"] = hrflow_profile.get("reference")
    return breezy_profile


def format_date_to_iso(date):
    year = date.get("year")
    month = date.get("month")
    day = date.get("day")
    # Check if the date is complete, i.e., year, month, and day are all present
    if year is not None and month is not None and day is not None:
        return f"{year:04d}-{month:02d}-{day:02d}"
    elif year is not None and month is not None:
        return f"{year:04d}-{month:02d}"
    elif year is not None:
        return f"{year:04d}"
    else:
        return None


def format_candidate(breezy_profile: t.Dict) -> t.Dict:
    """
    Format a Breezy profile object into a Hrflow profile object
    Args:
        data (BreezyProfileModel): Breezy Profile to format
    Returns:
        HrFlowProfile: a Hrflow formatted profile object
    """
    info = dict(
        full_name=breezy_profile["name"],
        first_name=breezy_profile["name"].split(" ")[0],
        last_name=breezy_profile["name"].split(" ")[-1],
        email=breezy_profile["email_address"],
        phone=breezy_profile["phone_number"],
        urls=[
            dict(type=social_profile["type"], url=social_profile["url"])
            for social_profile in breezy_profile.get("social_profiles", [])
        ],
        summary=breezy_profile["summary"],
        location={"text": breezy_profile.get("address", ""), "lat": None, "lng": None},
    )

    educations = []
    for education in breezy_profile.get("education", []):
        formatted_education = dict()
        formatted_education["school"] = education.get("school_name")
        degree = education.get("degree")
        field_of_study = education.get("field_of_study")
        if degree or field_of_study:
            formatted_education["title"] = (
                (degree if degree else "")
                + " "
                + (field_of_study if field_of_study else "")
            ).strip()
        formatted_education["description"] = (
            formatted_education["title"] + " at " + formatted_education["school"]
        )
        formatted_education["date_start"] = None
        formatted_education["date_end"] = None
        if education.get("start_date") is not None:
            formatted_education["date_start"] = format_date_to_iso(
                education["start_date"]
            )
        if education.get("end_date") is not None:
            formatted_education["date_end"] = format_date_to_iso(education["end_date"])
        formatted_education["location"] = {"text": None, "lat": None, "lng": None}
        educations.append(formatted_education)

    experiences = []
    for experience in breezy_profile.get("work_history", []):
        formatted_experience = dict()
        formatted_experience["company"] = experience.get("company_name")

        formatted_experience["title"] = experience.get("title")
        formatted_experience["description"] = experience.get("summary")
        formatted_experience["date_start"] = None
        formatted_experience["date_end"] = None
        if experience.get("start_date") is not None:
            formatted_experience["date_start"] = format_date_to_iso(
                experience["start_date"]
            )
        if experience.get("end_date") is not None:
            formatted_experience["date_end"] = format_date_to_iso(
                experience["end_date"]
            )
        formatted_experience["location"] = {"text": None, "lat": None, "lng": None}
        experiences.append(formatted_experience)

    tags = []
    t = lambda name, value: dict(name=name, value=value)
    tags.append(t("breezy_hr_tags", breezy_profile.get("tags")))
    tags.append(t("headline", breezy_profile.get("headline")))
    tags.append(t("origin", breezy_profile.get("origin")))
    tags.append(t("source", breezy_profile.get("source", {}).get("name")))
    tags.append(t("sourced_by", breezy_profile.get("sourced_by")))
    tags.append(t("stage", breezy_profile.get("stage", {}).get("name")))
    tags.append(
        t("overall_score", breezy_profile.get("overall_score", {}).get("average_score"))
    )
    hrflow_profile = dict(
        reference=breezy_profile.get("_id"),
        info=info,
        created_at=breezy_profile.get("creation_date"),
        updated_at=breezy_profile.get("updated_date"),
        experiences=experiences,
        educations=educations,
        skills=[],
        tags=tags,
        resume=breezy_profile.get("resume"),
    )

    return hrflow_profile


BreezyHR = Connector(
    name="Breezy HR",
    type=ConnectorType.ATS,
    subtype="breezyhr",
    description=(
        "Breezyhr is an end-to-end recruiting software "
        "to help you attract & hire great employees with less effort"
    ),
    url="https://breezy.hr/",
    warehouse=BreezyHrWarehouse,
    flows=(
        Flow(Mode.create, Entity.job, Direction.inbound, format=format_job),
        Flow(Mode.update, Entity.job, Direction.inbound, format=format_job),
        Flow(
            Mode.archive, Entity.job, Direction.inbound, format=format_archive_in_hrflow
        ),
        Flow(Mode.create, Entity.profile, Direction.inbound, format=format_candidate),
        Flow(Mode.update, Entity.profile, Direction.inbound, format=format_candidate),
        Flow(Mode.create, Entity.profile, Direction.outbound, format=format_profile),
        Flow(
            Mode.update,
            Entity.profile,
            Direction.outbound,
            format=format_profile_for_update,
        ),
        Flow(
            Mode.archive,
            Entity.profile,
            Direction.inbound,
            format=format_archive_in_hrflow,
        ),
    ),
)



import typing as t
from enum import Enum
from logging import LoggerAdapter

import requests
from msgspec import Meta, Struct
from msgspec.structs import asdict
from typing_extensions import Annotated

from hrflow_connectors.v2.connectors.adzuna.schemas import AdzunaJob
from hrflow_connectors.v2.core.common import Entity
from hrflow_connectors.v2.core.warehouse import (
    Aisle,
    Criterias,
    Endpoint,
    Endpoints,
    ReadOperation,
    merge,
)

ADZUNA_BASE_URL = "http://api.adzuna.com/v1/api"
SEARCH_JOBS_ENDPOINT = Endpoint(
    name="Get Adzuna jobs",
    description="Use this endpoint to retrieve Adzuna's job advertisement listings.",
    url="https://api.adzuna.com/v1/doc/Search.md",
)

MAX_NUMBER_OF_PAGES = 10


class CountryCode(str, Enum):
    GB = "gb"
    AT = "at"
    AU = "au"
    BR = "br"
    CA = "ca"
    DE = "de"
    FR = "fr"
    IN = "in"
    IT = "it"
    NL = "nl"
    NZ = "nz"
    PL = "pl"
    RU = "ru"
    SG = "sg"
    US = "us"
    ZA = "za"


class SortDir(str, Enum):
    UP = "up"
    DOWN = "down"


class SortKey(str, Enum):
    DEFAULT = "default"
    HYBRID = "hybrid"
    DATE = "date"
    SALARY = "salary"
    RELEVANCE = "relevance"


class Filter(str, Enum):
    ON = "1"


class AuthParameters(Struct):
    app_id: Annotated[
        t.Optional[str],
        Meta(
            description="Application ID, supplied by Adzuna",
        ),
    ]
    app_key: Annotated[
        t.Optional[str],
        Meta(
            description="Application key, supplied by Adzuna",
        ),
    ]


class ReadJobsParameters(Struct):
    country: Annotated[
        CountryCode,
        Meta(
            description="ISO 8601 country code of the country of interest",
        ),
    ]
    results_per_page: Annotated[
        t.Optional[int],
        Meta(
            description="The number of results to include on a page of search results.",
        ),
    ] = 50
    what: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "The keywords to search for. Multiple terms may be space separated."
            ),
        ),
    ] = None
    what_and: Annotated[
        t.Optional[str],
        Meta(
            description="The keywords to search for, all keywords must be found.",
        ),
    ] = None
    what_phrase: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "An entire phrase which must be found in the description or title."
            ),
        ),
    ] = None
    what_or: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "The keywords to search for, any keywords may be found. Multiple terms"
                " may be space separated."
            ),
        ),
    ] = None
    what_exclude: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "Keywords to exclude from the search. Multiple terms may be space"
                " separated."
            ),
        ),
    ] = None
    title_only: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "Keywords to find, but only in the title. Multiple terms may be space"
                " separated."
            ),
        ),
    ] = None
    where: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "The geographic centre of the search. Place names, postal codes, etc."
                " may be used.	"
            ),
        ),
    ] = None
    distance: Annotated[
        t.Optional[int],
        Meta(
            description=(
                "The distance in kilometres from the centre of the place described by"
                " the 'where' parameter. Defaults to 5km."
            ),
        ),
    ] = None
    location0: Annotated[
        t.Optional[str],
        Meta(
            description=(
                """The location fields may be used to describe a location,"""
                """ in a similar form to that returned in """
                """a Adzuna::API::Response::Location object.For example,"""
                """ "location0=UK&location1=South East England&location2=Surrey" """
                """ will performn a search over the county of Surrey."""
            ),
        ),
    ] = None
    location1: t.Optional[str] = None

    location2: t.Optional[str] = None

    location3: t.Optional[str] = None

    location4: t.Optional[str] = None

    location5: t.Optional[str] = None

    location6: t.Optional[str] = None

    location7: t.Optional[str] = None

    max_days_old: Annotated[
        t.Optional[int],
        Meta(
            description=(
                "The age of the oldest advertisment in days that will be returned."
            ),
        ),
    ] = None
    category: Annotated[
        t.Optional[str],
        Meta(
            description='The category tag, as returned by the "category" endpoint.',
        ),
    ] = None
    sort_dir: Annotated[
        t.Optional[SortDir],
        Meta(
            description="The direction to order the search results.",
        ),
    ] = None
    sort_by: Annotated[
        t.Optional[SortKey],
        Meta(
            description="The ordering of the search results.",
        ),
    ] = None
    salary_min: Annotated[
        t.Optional[int],
        Meta(
            description="The minimum salary we wish to get results for.",
        ),
    ] = None
    salary_max: Annotated[
        t.Optional[int],
        Meta(
            description="The maximum salary we wish to get results for.",
        ),
    ] = None
    salary_include_unknown: Annotated[
        t.Optional[Filter],
        Meta(
            description="""If set it "1", jobs without a known salary are returned.""",
        ),
    ] = None
    full_time: Annotated[
        t.Optional[Filter],
        Meta(
            description="""If set to "1", only full time jobs will be returned.""",
        ),
    ] = None
    part_time: Annotated[
        t.Optional[Filter],
        Meta(
            description="""If set to "1", only part time jobs will be returned.""",
        ),
    ] = None
    contract: Annotated[
        t.Optional[Filter],
        Meta(
            description="""If set to "1", only contract jobs will be returned.""",
        ),
    ] = None
    permanent: Annotated[
        t.Optional[Filter],
        Meta(
            description="""If set to "1", only permanent jobs will be returned.""",
        ),
    ] = None
    company: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "The canonical company name. This may be returned in a"
                " Adzuna::API::Response::Company object when a job is returned. A full"
                " list of allowed terms in not available through the API."
            ),
        ),
    ] = None


def read(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: ReadJobsParameters,
    incremental: bool,
    incremental_token: t.Optional[str],
) -> t.Iterable[t.Dict]:
    params = asdict(parameters)
    page = 1
    country = params.pop("country")
    params["app_id"] = auth_parameters.app_id
    params["app_key"] = auth_parameters.app_key

    while True:
        ADZUNA_JOBS_SEARCH_ENDPOINT = "{}/jobs/{}/search/{}".format(
            ADZUNA_BASE_URL, country.value, page
        )
        response = requests.get(
            ADZUNA_JOBS_SEARCH_ENDPOINT,
            params=params,
        )
        if response.status_code // 100 != 2:
            adapter.error(
                "Failed to pull jobs from Adzuna params={}"
                " status_code={} response={}".format(
                    params, response.status_code, response.text
                )
            )
            raise Exception("Failed to pull jobs from Adzuna")

        jobs = response.json()["results"]
        for job in jobs:
            yield job

        page += 1
        if len(jobs) == 0 or page == MAX_NUMBER_OF_PAGES:
            break


JobsAisle = Aisle(
    name=Entity.job,
    schema=AdzunaJob,
    read=ReadOperation(
        criterias=Criterias(
            create=ReadJobsParameters,
            update=ReadJobsParameters,
            archive=ReadJobsParameters,
        ),
        function=merge(create=read, update=read, archive=read),
        endpoints=Endpoints(
            create=SEARCH_JOBS_ENDPOINT,
            update=SEARCH_JOBS_ENDPOINT,
            archive=SEARCH_JOBS_ENDPOINT,
        ),
    ),
)



import typing as t
from enum import Enum
from logging import LoggerAdapter

import requests
from msgspec import Meta, Struct
from msgspec import json as msgspec_json
from typing_extensions import Annotated

from hrflow_connectors.v2.connectors.flatchr.schemas import (
    FlatchrProfile,
    FlatchrVacancy,
)
from hrflow_connectors.v2.core.common import Entity
from hrflow_connectors.v2.core.warehouse import (
    Aisle,
    Criterias,
    Endpoint,
    Endpoints,
    ReadOperation,
    WriteOperation,
    merge,
)

FLATCHR_CAREERS_BASE_URL = "https://careers.flatchr.io"


class FlatchrBaseURL(str, Enum):
    PRODUCTION = "https://api.flatchr.io/"
    TESTING = "https://api.demo.flatchr.io"


class AuthParameters(Struct):
    api_key: Annotated[
        str,
        Meta(
            description="The API key to authenticate with the Flatchr API",
        ),
    ]
    company_id: Annotated[
        str,
        Meta(
            description="The ID of the company to authenticate with",
        ),
    ]
    env_base_url: Annotated[
        FlatchrBaseURL,
        Meta(
            description="The base URL of the Flatchr API",
        ),
    ] = FlatchrBaseURL.PRODUCTION


class ReadProfilesParameters(Struct, omit_defaults=True):
    firstname: Annotated[
        t.Optional[str],
        Meta(
            description="The firstname of the candidate to search for",
        ),
    ] = None
    lastname: Annotated[
        t.Optional[str],
        Meta(
            description="The lastname of the candidate to search for",
        ),
    ] = None
    email: Annotated[
        t.Optional[str],
        Meta(
            description="The email of the candidate to search for",
        ),
    ] = None
    hired: Annotated[
        t.Optional[bool],
        Meta(
            description="Whether the candidate has been hired or not",
        ),
    ] = None
    column: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "The column in which the candidate is located, Ex: 'Entretien RH'"
            ),
        ),
    ] = None
    start: Annotated[
        t.Optional[str],
        Meta(
            description="The start date in MM/DD/YY of the search",
        ),
    ] = None
    end: Annotated[
        t.Optional[str],
        Meta(
            description="The end date in MM/DD/YY of the search",
        ),
    ] = None
    vacancy: Annotated[
        t.Optional[str],
        Meta(description="id of the offer in which the candidate is involved"),
    ] = None


class WriteProfilesParameters(Struct):
    vacancy_slug: Annotated[
        str,
        Meta(
            description="The slug of the offer to assign the candidate to",
        ),
    ]


class UpdateProfilesParameters(Struct):
    app_name: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "Name of the application inserting the data (e.g. if used by a CV"
                " parsing application)\nExemple: 'HRMatch'"
            ),
        ),
    ] = None


class ArchiveProfilesParameters(Struct):
    vacancy_id: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "The ID of the offer to assign the candidate to\nEquivalent to id in"
                " the Flatchr API not vacancy_id nor the slug"
            ),
        ),
    ] = None


class ReadJobsParameters(Struct):
    pass


def read_profiles(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: ReadProfilesParameters,
    incremental: bool,
    incremental_token: t.Optional[str],
) -> t.Iterable[t.Dict]:
    FLATCHR_CANDIDATES_ENDPOINT = (
        "{base_ur}/company/{companyId}/search/applicants".format(
            base_ur=auth_parameters.env_base_url, companyId=auth_parameters.company_id
        )
    )
    headers = {"Authorization": "Bearer {}".format(auth_parameters.api_key)}
    data = msgspec_json.decode(msgspec_json.encode(parameters), type=dict)

    response = requests.post(
        url=FLATCHR_CANDIDATES_ENDPOINT, headers=headers, json=data
    )
    if response.status_code // 100 != 2:
        adapter.error(
            "Failed to fetch profiles from Flatchr status_code={} response={}".format(
                response.status_code, response.text
            )
        )
        raise Exception("Failed to fetch profiles from Flatchr")
    profiles = response.json()
    # The following steps are a workaround around limitations and misconfigurations in the Flatchr API
    # We go through the collumns endpoint to get the id of the column the candidate is in right now using the column name
    columns_response = requests.get(
        "{base_url}/company/{companyId}/columns".format(
            base_url=auth_parameters.env_base_url, companyId=auth_parameters.company_id
        ),
        headers=headers,
    )
    columns = columns_response.json()
    # we go through the vacancies endpoint to get the id of the vacancy the candidate is attached to using the vacancy_id in the profile
    vacancies_response = requests.get(
        "{base_url}/company/{companyId}/vacancies".format(
            base_url=auth_parameters.env_base_url, companyId=auth_parameters.company_id
        ),
        headers=headers,
    )
    vacancies = vacancies_response.json()

    for profile in profiles:
        column_id = next(
            (
                column["id"]
                for column in columns
                if column["title"] == profile["column"]
            ),
            None,
        )
        vacancy_id = next(
            (
                vacancy["id"]
                for vacancy in vacancies
                if vacancy["vacancy_id"] == profile["vacancy_id"]
            ),
            None,
        )
        # We move candidate to the same column he's in right now just to get the candidate id
        if column_id and vacancy_id:
            move_candidate_response = requests.put(
                "{base_url}/company/{companyId}/vacancy/{vacancyId}/applicant/{applicantId}"
                .format(
                    base_url=auth_parameters.env_base_url,
                    companyId=auth_parameters.company_id,
                    vacancyId=vacancy_id,
                    applicantId=profile["applicant"],
                ),
                headers=headers,
                json={"column_id": column_id},
            )
            if move_candidate_response.status_code // 100 != 2:
                adapter.error(
                    "Failed to move candidate in Flatchr status_code={} response={}"
                    .format(
                        move_candidate_response.status_code,
                        move_candidate_response.text,
                    )
                )
            candidate_additional_info = move_candidate_response.json()["candidate"]
            # Get the candidate resume
            candidate_resume_response = requests.get(
                "{base_url}/company/{companyId}/files/expire?key={key}&extension={extension}&token={token}"
                .format(
                    base_url=auth_parameters.env_base_url,
                    companyId=auth_parameters.company_id,
                    key=candidate_additional_info["aws"]["key"],
                    extension=candidate_additional_info["aws"]["extension"],
                    token=auth_parameters.api_key,
                )
            )
            if candidate_resume_response.status_code // 100 != 2:
                adapter.error(
                    "Failed to get candidate resume in Flatchr status_code={}"
                    " response={}".format(
                        candidate_resume_response.status_code,
                        candidate_resume_response.text,
                    )
                )
                candidate_additional_info["resume"] = None
            else:
                candidate_additional_info["resume"] = {
                    "raw": candidate_resume_response.content,
                    "content_type": candidate_additional_info["aws"]["extension"],
                }
            profile.update(candidate_additional_info)
            yield profile
        else:
            adapter.error(
                "Failed to get column id or vacancy id for candidate column_id={} "
                "vacancy_id={}".format(column_id, vacancy_id)
            )
            raise Exception("Failed to get column id or vacancy id for candidate")


def write(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: WriteProfilesParameters,
    items: t.Iterable[dict],
) -> t.List[t.Dict]:
    failed_profiles = []
    FLATCHR_POST_CANDIDATES_ENDPOINT = "{base_url}/vacancy/candidate/json".format(
        base_url=FLATCHR_CAREERS_BASE_URL
    )

    headers = {"Authorization": "Bearer {}".format(auth_parameters.api_key)}

    for profile in items:
        profile_creation_json = dict(
            **profile, vacancy=parameters.vacancy_slug, token=auth_parameters.api_key
        )
        response = requests.post(
            FLATCHR_POST_CANDIDATES_ENDPOINT,
            headers=headers,
            json=profile_creation_json,
        )
        if response.status_code // 100 != 2:
            adapter.error(
                "Failed to create profile in Flatchr status_code={} response={}".format(
                    response.status_code, response.text
                )
            )
            failed_profiles.append(profile)

    return failed_profiles


def archive(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: ArchiveProfilesParameters,
    items: t.Iterable[dict],
) -> t.List[t.Dict]:
    failed_profiles = []
    headers = {"Authorization": "Bearer {}".format(auth_parameters.api_key)}

    for profile in items:
        if parameters.vacancy_id:
            vacancy_id = parameters.vacancy_id
        else:
            # we go through the vacancies endpoint to get the id of the vacancy the candidate is attached to using the vacancy_id in the profile
            vacancies_response = requests.get(
                "{base_url}/company/{companyId}/vacancies".format(
                    base_url=auth_parameters.env_base_url,
                    companyId=auth_parameters.company_id,
                ),
                headers=headers,
            )
            vacancies = vacancies_response.json()
            vacancy_id = next(
                (
                    vacancy["id"]
                    for vacancy in vacancies
                    if vacancy["vacancy_id"] == profile["vacancy_id"]
                ),
                None,
            )

        archive_candidate_response = requests.delete(
            "{base_url}/company/{companyId}/vacancy/{vacancyId}/applicant/{applicantId}"
            .format(
                base_url=auth_parameters.env_base_url,
                companyId=auth_parameters.company_id,
                vacancyId=vacancy_id,
                applicantId=profile["applicant_id"],
            ),
            headers=headers,
        )
        if archive_candidate_response.status_code // 100 != 2:
            adapter.error(
                "Failed to archive profile in Flatchr status_code={} response={}"
                .format(
                    archive_candidate_response.status_code,
                    archive_candidate_response.text,
                )
            )
            failed_profiles.append(profile)

    return failed_profiles


def update(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: UpdateProfilesParameters,
    items: t.Iterable[dict],
) -> t.List[t.Dict]:
    failed_profiles = []
    FLATCHR_UPDATE_CANDIDATES_ENDPOINT = (
        "{base_url}/company/{companyID}/search/candidate".format(
            base_url=auth_parameters.env_base_url, companyID=auth_parameters.company_id
        )
    )
    headers = {"Authorization": "Bearer {}".format(auth_parameters.api_key)}

    for profile in items:
        json_data = dict(reference=profile.pop("id"), type="applicants", value=profile)
        response = requests.post(
            url=FLATCHR_UPDATE_CANDIDATES_ENDPOINT,
            headers=headers,
            json=json_data,
        )
        if response.status_code // 100 != 2:
            adapter.error(
                "Failed to update profile in Flatchr status_code={} response={}".format(
                    response.status_code, response.text
                )
            )
            failed_profiles.append(profile)

    return failed_profiles


def read_jobs(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: ReadProfilesParameters,
    incremental: bool,
    incremental_token: t.Optional[str],
) -> t.Iterable[t.Dict]:
    FLATCHR_VACANCIES_ENDPOINT = "{base_ur}/company/{companyId}/vacancies".format(
        base_ur=auth_parameters.env_base_url, companyId=auth_parameters.company_id
    )
    headers = {"Authorization": "Bearer {}".format(auth_parameters.api_key)}
    response = requests.get(url=FLATCHR_VACANCIES_ENDPOINT, headers=headers)

    if response.status_code // 100 != 2:
        adapter.error(
            "Failed to fetch jobs from Flatchr status_code={} response={}".format(
                response.status_code, response.text
            )
        )
        raise Exception("Failed to fetch jobs from Flatchr")
    jobs = response.json()
    for job in jobs:
        yield job


JobsAisle = Aisle(
    name=Entity.job,
    schema=FlatchrVacancy,
    read=ReadOperation(
        criterias=Criterias(
            create=ReadJobsParameters,
            update=ReadJobsParameters,
            archive=ReadJobsParameters,
        ),
        function=merge(create=read_jobs, update=read_jobs, archive=read_jobs),
        endpoints=Endpoints(
            create=Endpoint(
                name="Get Vacancies",
                description="search for one or more vacancies",
                url="https://developers.flatchr.io/docs/QuickStart/Annonces/Recuperer_l_ensemble_de_vos_annonces",
            ),
        ),
    ),
)
ProfilesAisle = Aisle(
    name=Entity.profile,
    schema=FlatchrProfile,
    read=ReadOperation(
        criterias=Criterias(
            create=ReadProfilesParameters,
            update=ReadProfilesParameters,
            archive=ReadProfilesParameters,
        ),
        function=merge(
            create=read_profiles, update=read_profiles, archive=read_profiles
        ),
        endpoints=Endpoints(
            create=Endpoint(
                name="Get Candidates",
                description="search for one or more candidates",
                url="https://developers.flatchr.io/docs/QuickStart/Candidats/Recuperer_un_candidat",
            ),
        ),
    ),
    write=WriteOperation(
        criterias=Criterias(
            create=WriteProfilesParameters,
            update=UpdateProfilesParameters,
            archive=ArchiveProfilesParameters,
        ),
        function=merge(create=write, update=update, archive=archive),
        endpoints=Endpoints(
            create=Endpoint(
                name="Create Candidate",
                description="Create a candidate",
                url="https://developers.flatchr.io/docs/QuickStart/Candidats/Creer_un_candidat",
            ),
            update=Endpoint(
                name="Update Candidate",
                description="Update a candidate",
                url="https://developers.flatchr.io/docs/QuickStart/Candidats/Meta_informations_candidat",
            ),
            archive=Endpoint(
                name="Archive Candidate",
                description="Archive a candidate",
                url="https://developers.flatchr.io/docs/QuickStart/Candidats/Archiver_un_candidat",
            ),
        ),
    ),
)



import typing as t
from enum import Enum

from msgspec import Struct, field


class ZohoUser(Struct):
    name: str
    id: str


class ZohoRecruiter(Struct):
    name: str
    id: str
    email: str
    photoSrc: str


class ZohoState(str, Enum):
    DRAFT = "draft"
    SAVE = "save"


class Approval(Struct):
    delegate: bool
    approve: bool
    reject: bool
    resubmit: bool


class JobOpening(Struct):
    id: str
    Posting_Title: t.Optional[str]
    Client_Name: str
    Job_Opening_Name: str
    Job_Description: t.Optional[str]
    Job_Type: t.Optional[str]
    Job_Opening_Status: t.Optional[str]
    Job_Opening_ID: t.Optional[str]
    currency_symbol: t.Optional[str] = field(name="$currency_symbol")
    Required_Skills: t.Optional[str]
    Industry: t.Optional[str]
    state: t.Optional[ZohoState] = field(name="$state")
    process_flow: t.Optional[bool] = field(name="$process_flow")
    City: t.Optional[str]
    State: t.Optional[str]
    Zip_Code: t.Optional[str]
    Country: t.Optional[str]
    approved: t.Optional[bool] = field(name="$approved")
    Created_Time: t.Optional[str]
    Modified_Time: t.Optional[str]
    Date_Opened: t.Optional[str]
    Target_Date: t.Optional[str]
    Last_Activity_Time: t.Optional[str]
    approval: t.Optional[Approval] = field(name="$approval")
    approval_state: t.Optional[str] = field(name="$approval_state")
    Remote_Job: t.Optional[bool]
    followed: t.Optional[bool] = field(name="$followed")
    editable: t.Optional[bool] = field(name="$editable")
    Is_Locked: t.Optional[bool]
    Salary: t.Optional[str]
    Work_Experience: t.Optional[str]
    Number_of_Positions: t.Optional[str]
    Associated_Tags: t.Optional[t.List[str]]
    Account_Manager: t.Optional[ZohoUser]
    Assigned_Recruiter: t.Optional[t.List[ZohoRecruiter]]
    Assigned_Recruiters: t.Optional[t.List[ZohoRecruiter]]
    Contact_Name: t.Optional[ZohoUser]
    Created_By: t.Optional[ZohoUser]
    Modified_By: t.Optional[ZohoUser]
    Is_Attachment_Present: t.Optional[bool]
    Keep_on_Career_Site: t.Optional[bool]
    No_of_Candidates_Hired: t.Optional[int]
    Expected_Revenue: t.Optional[float]
    Is_Hot_Job_Opening: t.Optional[bool]
    Publish: t.Optional[bool]
    Actual_Revenue: t.Optional[float]
    Missed_Revenue: t.Optional[float]
    No_of_Candidates_Associated: t.Optional[int]
    Revenue_per_Position: t.Optional[float]


class ZohoDuration(Struct):
    from_: str = field(name="from")
    to: t.Optional[str]


class Experience_Detail(Struct):
    id: str
    Company: str
    I_currently_work_here: bool
    Summary: str
    Work_Duration: ZohoDuration
    Occupation_Title: str


class Educational_Detail(Struct):
    id: str
    Institute_School: str
    Currently_pursuing: bool
    Degree: str
    Major_Department: str
    Duration: ZohoDuration


class Candidate(Struct):
    id: str
    First_Name: t.Optional[str]
    Last_Name: str
    Full_Name: t.Optional[str]
    Salutation: t.Optional[str]
    Email: t.Optional[str]
    Phone: t.Optional[str]
    Mobile: t.Optional[str]
    Fax: t.Optional[str]
    Street: t.Optional[str]
    City: t.Optional[str]
    State: t.Optional[str]
    Zip_Code: t.Optional[str]
    Country: t.Optional[str]
    Current_Job_Title: t.Optional[str]
    Current_Salary: t.Optional[int]
    Expected_Salary: t.Optional[int]
    Current_Employer: t.Optional[str]
    Experience_in_Years: t.Optional[int]
    Experience_Details: t.Optional[t.List[Experience_Detail]]
    Educational_Details: t.Optional[t.List[Educational_Detail]]
    Highest_Qualification_Held: t.Optional[str]
    Skill_Set: t.Optional[str]
    Created_Time: t.Optional[str]
    Last_Activity_Time: t.Optional[str]
    Updated_On: t.Optional[str]
    Last_Mailed_Time: t.Optional[str]
    Source: t.Optional[str]
    Origin: t.Optional[str]
    currency_symbol: t.Optional[str] = field(name="$currency_symbol")
    whatsapp_available: t.Optional[bool] = field(name="$whatsapp_available")
    state: t.Optional[ZohoState] = field(name="$state")
    converted: t.Optional[bool] = field(name="$converted")
    process_flow: t.Optional[bool] = field(name="$process_flow")
    approved: t.Optional[bool] = field(name="$approved")
    approval: t.Optional[str] = field(name="$approval")
    approval_state: t.Optional[str] = field(name="$approval_state")
    Candidate_Status: t.Optional[str]
    Candidate_ID: t.Optional[str]
    Candidate_Owner: t.Optional[ZohoUser]
    followed: t.Optional[bool] = field(name="$followed")
    LinkedIn__s: t.Optional[str]
    editable: t.Optional[bool] = field(name="$editable")
    Is_Locked: t.Optional[bool]
    Is_Unqualified: t.Optional[bool]
    Associated_Tags: t.Optional[t.List[str]]
    Additional_Info: t.Optional[str]
    Created_By: t.Optional[ZohoUser]
    Modified_By: t.Optional[ZohoUser]
    Secondary_Email: t.Optional[str]
    Is_Attachment_Present: t.Optional[bool]
    Rating: t.Optional[int]
    applied_with_linkedin: t.Optional[bool] = field(name="$applied_with_linkedin")
    No_of_Applications: t.Optional[int]
    Website: t.Optional[str]
    Twitter: t.Optional[str]
    Facebook__s: t.Optional[str]
    social_profiles: t.Optional[str] = field(name="$social_profiles")
    Skype_ID: t.Optional[str]
    Candidate_Stage: t.Optional[str]
    Fresh_Candidate: t.Optional[bool]
    Email_Opt_Out: t.Optional[bool]
    converted_detail: t.Optional[t.Dict] = field(name="$converted_detail")
    Career_Page_Invite_Status: t.Optional[str]
    Associated_any_Social_Profiles: t.Optional[bool]



from hrflow_connectors.v2.connectors.admen.aisles import (
    AuthParameters,
    JobsAisle,
    ProfilesAisle,
)
from hrflow_connectors.v2.core.warehouse import Warehouse

AdmenWarehouse = Warehouse(auth=AuthParameters, aisles=(JobsAisle, ProfilesAisle))



import re


def remove_html_tags(text: str) -> str:
    """
    Remove all HTML tags in a string
    Args:
        text (str): text to clean
    Returns:
        str: cleaned text (without HTML tags)
    """
    return re.sub("<[^<]+?>", "", text)


def is_valid_url(url):
    regex = r"^https?:\/\/(www\.)?[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}\/?[^\s]*$"
    return re.match(regex, url) is not None



import typing as t
from dataclasses import dataclass, field
from logging import LoggerAdapter

from msgspec import Struct

from hrflow_connectors.v2.core.common import Entity, Mode, Parameters, Schema


class SpecificRead(t.Protocol):
    def __call__(
        self,
        adapter: LoggerAdapter,
        auth_parameters: t.Any,
        parameters: t.Any,
        incremental: bool,
        incremental_token: t.Optional[str],
    ) -> t.Iterable[dict]:
        ...  # pragma: nocover


class Read(t.Protocol):
    def __call__(
        self,
        *,
        mode: Mode,
        adapter: LoggerAdapter,
        auth_parameters: t.Any,
        parameters: t.Any,
        incremental: bool,
        incremental_token: t.Optional[str],
    ) -> t.Iterable[dict]:
        ...  # pragma: nocover


class SpecificWrite(t.Protocol):
    def __call__(
        self,
        adapter: LoggerAdapter,
        auth_parameters: t.Any,
        parameters: t.Any,
        items: t.Iterable[dict],
    ) -> list[dict]:
        ...  # pragma: nocover


class Write(t.Protocol):
    def __call__(
        self,
        *,
        mode: Mode,
        adapter: LoggerAdapter,
        auth_parameters: t.Any,
        parameters: t.Any,
        items: t.Iterable[dict],
    ) -> list[dict]:
        ...  # pragma: nocover


class ModeIsNotSupported(Exception):
    pass


@t.overload
def merge(
    *,
    create: t.Optional[SpecificRead] = None,
    update: t.Optional[SpecificRead] = None,
    archive: t.Optional[SpecificRead] = None,
) -> Read:
    ...  # pragma: nocover


@t.overload
def merge(
    *,
    create: t.Optional[SpecificWrite] = None,
    update: t.Optional[SpecificWrite] = None,
    archive: t.Optional[SpecificWrite] = None,
) -> Write:
    ...  # pragma: nocover


def merge(
    *,
    create: t.Optional[t.Callable[..., t.Any]] = None,
    update: t.Optional[t.Callable[..., t.Any]] = None,
    archive: t.Optional[t.Callable[..., t.Any]] = None,
) -> t.Callable[..., t.Any]:
    def merged(mode: Mode, **kwargs: t.Any) -> list[dict]:
        if mode is Mode.create and create is not None:
            return create(**kwargs)

        if mode is Mode.update and update is not None:
            return update(**kwargs)

        if mode is Mode.archive and archive is not None:
            return archive(**kwargs)

        raise ModeIsNotSupported(f"{mode} mode is not supported")

    return merged


@dataclass
class Endpoint:
    name: str
    description: str
    url: str


@dataclass
class Criterias:
    create: t.Optional[Schema] = None
    update: t.Optional[Schema] = None
    archive: t.Optional[Schema] = None

    def parameters(self, mode: Mode):
        if mode is Mode.create:
            return self.create
        if mode is Mode.update:
            return self.update
        if mode is Mode.archive:
            return self.archive


@dataclass
class Endpoints:
    create: t.Optional[Endpoint] = None
    update: t.Optional[Endpoint] = None
    archive: t.Optional[Endpoint] = None

    def for_mode(self, mode: Mode):
        if mode is Mode.create:
            return self.create
        if mode is Mode.update:
            return self.update
        if mode is Mode.archive:
            return self.archive


@dataclass
class IncrementalTokenHandler:
    create: t.Optional[t.Callable[[dict], str]] = None
    update: t.Optional[t.Callable[[dict], str]] = None
    archive: t.Optional[t.Callable[[dict], str]] = None

    def get_handler(self, mode: Mode) -> t.Optional[t.Callable[[dict], str]]:
        if mode is Mode.create:
            return self.create
        if mode is Mode.update:
            return self.update
        if mode is Mode.archive:
            return self.archive

    def __call__(self, mode: Mode) -> t.Optional[t.Callable[[dict], str]]:
        return self.get_handler(mode)


OperationT = t.TypeVar("OperationT", Read, Write)


@dataclass
class Operation(t.Generic[OperationT]):
    function: OperationT
    criterias: Criterias
    endpoints: Endpoints = field(default_factory=Endpoints)


@dataclass
class ReadOperation(Operation[Read]):
    # get_incremental_token: t.Optional[t.Callable[[dict], str]] = None
    incremental_token_handler: t.Optional[IncrementalTokenHandler] = None

    def supports_incremental(self, mode: Mode) -> bool:
        if self.incremental_token_handler is None:
            return False
        return self.incremental_token_handler.get_handler(mode) is not None

    def __call__(
        self,
        *,
        mode: Mode,
        adapter: LoggerAdapter,
        auth_parameters: Parameters,
        parameters: Parameters,
        incremental: bool,
        incremental_token: t.Optional[str],
    ):
        return self.function(
            mode=mode,
            adapter=adapter,
            auth_parameters=auth_parameters,
            parameters=parameters,
            incremental=incremental,
            incremental_token=incremental_token,
        )


@dataclass
class WriteOperation(Operation[Write]):
    def __call__(
        self,
        *,
        mode: Mode,
        adapter: LoggerAdapter,
        auth_parameters: Parameters,
        parameters: Parameters,
        items: list[dict],
    ):
        return self.function(
            mode=mode,
            adapter=adapter,
            auth_parameters=auth_parameters,
            parameters=parameters,
            items=items,
        )


@dataclass
class Aisle:
    name: Entity
    read: t.Optional[ReadOperation] = None
    write: t.Optional[WriteOperation] = None
    schema: Schema = field(default_factory=lambda: Struct)

    def parameters(self, operation: t.Literal["read", "write"], mode: Mode):
        if operation == "read" and self.read is not None:
            return self.read.criterias.parameters(mode)
        elif operation == "write" and self.write is not None:
            return self.write.criterias.parameters(mode)


@dataclass
class Warehouse:
    auth: Schema
    aisles: tuple[Aisle, ...]

    def get_aisle(self, entity: Entity) -> t.Optional[Aisle]:
        return self.__dict__.setdefault(
            "__aisle_by_entity__", {aisle.name: aisle for aisle in self.aisles}
        ).get(entity)



import typing as t

from hrflow_connectors.v2.connectors.francetravail.warehouse import (
    FranceTravailWarehouse,
)
from hrflow_connectors.v2.core.common import Direction, Entity, Mode
from hrflow_connectors.v2.core.connector import Connector, ConnectorType, Flow


def get_job_location(pole_emploi_location: t.Optional[t.Dict]) -> t.Dict:
    if pole_emploi_location is None:
        return dict(lat=None, lng=None, text="")

    return dict(
        lat=pole_emploi_location.get("latitude"),
        lng=pole_emploi_location.get("longitude"),
        text=" ".join(
            [
                value
                for field in ["libelle", "codePostal"]
                if (value := pole_emploi_location.get(field)) is not None
            ]
        ),
    )


def get_sections(job: t.Dict) -> t.List[t.Dict]:
    sections = []
    if "entreprise" in job and "description" in job["entreprise"]:
        sections.append(
            dict(
                name="company_description",
                title="Company Description",
                description=job["entreprise"]["description"],
            )
        )
    return sections


def get_requirements(pole_emploi_formations: t.List[t.Dict]) -> t.Optional[str]:
    requirements = "\n".join(
        [
            f"{formation['niveauLibelle']} en {formation['domaineLibelle']}"
            for formation in pole_emploi_formations
            if formation.get("niveauLibelle") and formation.get("domaineLibelle")
        ]
    )
    return requirements if requirements else None


def get_tags(job: t.Dict) -> t.List[t.Dict]:
    contact = job.get("contact", {})
    salary = job.get("salaire", {})
    tags = []

    # Helper function to create tag dictionaries
    create_tag = lambda name, value: dict(name=name, value=value)

    tags = [
        create_tag("rome_code", job.get("romeCode")),
        create_tag("rome_label", job.get("romeLibelle")),
        create_tag("contract_nature", job.get("natureContrat")),
        create_tag("contract_type", job.get("typeContratLibelle")),
        create_tag("accessible_to_disabled", job.get("accessibleTH")),
        create_tag("is_apprenticeship", job.get("alternance")),
        create_tag("experience_required", job.get("experienceExige") == "E"),
        create_tag("experience_description", job.get("experienceLibelle")),
        create_tag("salary_description", salary.get("libelle")),
        create_tag("working_hours_description", job.get("dureeTravailLibelle")),
        create_tag("working_hours_type", job.get("dureeTravailLibelleConverti")),
        create_tag("number_of_positions", job.get("nombrePostes")),
        create_tag("qualification_label", job.get("qualificationLibelle")),
        create_tag("sector_of_activity", job.get("secteurActiviteLibelle")),
        create_tag("company_name", job.get("entreprise", {}).get("nom")),
        create_tag("company_description", job.get("entreprise", {}).get("description")),
        create_tag("company_website", job.get("entreprise", {}).get("url")),
        create_tag("recruiter_name", contact.get("nom")),
        create_tag("recruiter_email", contact.get("courriel")),
        create_tag("recruiter_phone", contact.get("telephone")),
        create_tag("recruiter_website", contact.get("urlRecruteur")),
        create_tag("application_url", job.get("urlPostulation")),
    ]
    return tags


def format_job(
    pole_emploi_job: t.Dict,
) -> t.Dict:
    job = dict(
        name=pole_emploi_job.get("intitule"),
        reference=pole_emploi_job.get("id"),
        created_at=pole_emploi_job.get("dateCreation"),
        updated_at=pole_emploi_job.get("dateActualisation"),
        location=get_job_location(pole_emploi_job.get("lieuTravail")),
        url=pole_emploi_job.get("origineOffre", {}).get("urlOrigine"),
        summary=pole_emploi_job.get("description"),
        requirements=get_requirements(pole_emploi_job.get("formations", [])),
        skills=[
            dict(name=skill["libelle"], value=None)
            for skill in pole_emploi_job.get("competences", [])
        ],
        languages=[
            dict(name=language["libelle"], value=None)
            for language in pole_emploi_job.get("langues", [])
        ],
        sections=get_sections(pole_emploi_job),
        tags=get_tags(pole_emploi_job),
    )
    return job


DESCRIPTION = (
    "France Travail, formerly Pôle emploi until 31 December 2023, is a public"
    " administrative body responsible for employment in France. It was created on 19"
    " December 2008 from the merger of the Agence nationale pour l'emploi and the"
    " Association pour l'emploi dans l'industrie et le commerce."
)


def format_archive(
    pole_emploi_job: t.Dict,
) -> t.Dict:
    return dict(reference=pole_emploi_job.get("id"))


FranceTravail = Connector(
    name="France Travail (ex: Pole Emploi)",
    type=ConnectorType.JobBoard,
    subtype="francetravail",
    description=DESCRIPTION,
    url="https://www.francetravail.fr/",
    warehouse=FranceTravailWarehouse,
    flows=(
        Flow(Mode.create, Entity.job, Direction.inbound, format=format_job),
        Flow(Mode.update, Entity.job, Direction.inbound, format=format_job),
        Flow(Mode.archive, Entity.job, Direction.inbound, format=format_archive),
    ),
)



import typing as t

from msgspec import Meta, Struct, field
from typing_extensions import Annotated


class Location(Struct):
    text: Annotated[t.Optional[str], Meta(description="Location text address.")] = None
    lat: Annotated[
        t.Optional[float], Meta(description="Geocentric latitude of the Location.")
    ] = None
    lng: Annotated[
        t.Optional[float], Meta(description="Geocentric longitude of the Location.")
    ] = None
    fields: Annotated[
        t.Optional[dict[str, t.Any]],
        Meta(description="other location attributes like country, country_code etc"),
    ] = None


class GeneralEntitySchema(Struct):
    name: Annotated[str, Meta(description="Identification name of the Object")]
    value: Annotated[
        t.Optional[str], Meta(description="Value associated to the Object's name")
    ] = None


class Skill(Struct):
    name: Annotated[str, Meta(description="Identification name of the skill")]
    type: Annotated[
        t.Literal["hard", "soft"], Meta(description="Type of the skill. hard or soft")
    ]
    value: Annotated[
        t.Optional[str], Meta(description="Value associated to the skill")
    ] = None


class Label(Struct):
    board_key: Annotated[
        str,
        Meta(description="Identification key of the Board containing the target Job."),
    ]
    job_key: Annotated[str, Meta(description="Identification key of the Job.")]
    job_reference: Annotated[str, Meta(description="Custom identifier of the Job.")]
    stage: Annotated[
        t.Literal["yes", "no", "later"],
        Meta(
            description=(
                "Stage associated to the Profile following the action of a recruiter"
                " (yes, no, later)."
            )
        ),
    ]
    date_stage: Annotated[
        str,
        Meta(description="Date of the stage edit action. type: ('datetime ISO 8601')"),
    ]
    rating: Annotated[
        t.Optional[t.Literal[1, 2, 3, 4, 5]],
        Meta(
            description=(
                "Rating associated to the Profile following the action of a recruiter"
                " (from 1 to 5)."
            )
        ),
    ]
    date_rating: Annotated[
        str, Meta(description="Date of the rating action. type: ('datetime ISO 8601')")
    ]


# Job
class Section(Struct):
    name: Annotated[
        t.Optional[str],
        Meta(
            description="Identification name of a Section of the Job. Example: culture",
        ),
    ] = None
    title: Annotated[
        t.Optional[str],
        Meta(description="Display Title of a Section. Example: Corporate Culture"),
    ] = None
    description: Annotated[
        t.Optional[str],
        Meta(description="Text description of a Section: Example: Our values areNone"),
    ] = None


class RangesFloat(Struct):
    name: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "Identification name of a Range of floats attached "
                "to the Job. Example: salary"
            ),
        ),
    ] = None
    value_min: Annotated[
        t.Optional[float], Meta(description="Min value. Example: 500.")
    ] = None
    value_max: Annotated[
        t.Optional[float], Meta(description="Max value. Example: 100.")
    ] = None
    unit: Annotated[
        t.Optional[str], Meta(description="Unit of the value. Example: euros.")
    ] = None


class RangesDate(Struct):
    name: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "Identification name of a Range of dates attached"
                " to the Job. Example: availability."
            ),
        ),
    ] = None
    value_min: Annotated[
        t.Optional[str],
        Meta(description="Min value in datetime ISO 8601, Example: 500."),
    ] = None
    value_max: Annotated[
        t.Optional[str],
        Meta(description="Max value in datetime ISO 8601, Example: 1000"),
    ] = None


class HrFlowJob(Struct, kw_only=True):
    key: Annotated[
        t.Optional[str], Meta(description="Identification key of the Job.")
    ] = None
    reference: Annotated[
        t.Optional[str], Meta(description="Custom identifier of the Job.")
    ] = None
    name: Annotated[str, Meta(description="Job title.")]
    location: Annotated[Location, Meta(description="Job location object.")]
    sections: Annotated[list[Section], Meta(description="Job custom sections.")]
    url: Annotated[t.Optional[str], Meta(description="Job post original URL.")] = None
    summary: Annotated[
        t.Optional[str], Meta(description="Brief summary of the Job.")
    ] = None
    archieved_at: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "type: datetime ISO8601, Archive date of the Job. "
                "The value is null for unarchived Jobs."
            ),
        ),
    ] = None
    updated_at: Annotated[
        t.Optional[str],
        Meta(description="type: datetime ISO8601, Last update date of the Job."),
    ] = None
    created_at: Annotated[
        t.Optional[str],
        Meta(description="type: datetime ISO8601, Creation date of the Job."),
    ] = None
    skills: Annotated[
        t.Optional[list[Skill]], Meta(description="list of skills of the Job.")
    ] = None
    languages: Annotated[
        t.Optional[list[GeneralEntitySchema]],
        Meta(description="list of spoken languages of the Job"),
    ] = None
    certifications: Annotated[
        t.Optional[list[GeneralEntitySchema]],
        Meta(description="list of certifications of the Job."),
    ] = None
    courses: Annotated[
        t.Optional[list[GeneralEntitySchema]],
        Meta(description="list of courses of the Job"),
    ] = None
    tasks: Annotated[
        t.Optional[list[GeneralEntitySchema]],
        Meta(description="list of tasks of the Job"),
    ] = None
    tags: Annotated[
        t.Optional[list[GeneralEntitySchema]],
        Meta(description="list of tags of the Job"),
    ] = None
    metadatas: Annotated[
        t.Optional[list[GeneralEntitySchema]],
        Meta(description="list of metadatas of the Job"),
    ] = None
    ranges_float: Annotated[
        t.Optional[list[RangesFloat]], Meta(description="list of ranges of floats")
    ] = None
    ranges_date: Annotated[
        t.Optional[list[RangesDate]], Meta(description="list of ranges of dates")
    ] = None


# Profile
class InfoUrl(Struct):
    type: t.Literal["from_resume", "linkedin", "twitter", "facebook", "github"]
    url: t.Optional[str]


class ProfileInfo(Struct):
    full_name: t.Optional[str]
    first_name: t.Optional[str]
    last_name: t.Optional[str]
    email: t.Optional[str]
    phone: t.Optional[str]
    date_birth: Annotated[
        t.Optional[str], Meta(description="Profile date of birth")
    ] = None
    location: Annotated[
        t.Optional[Location], Meta(description="Profile location object")
    ] = None
    urls: Annotated[
        t.Optional[list[InfoUrl]],
        Meta(description="Profile social networks and URLs"),
    ] = None
    picture: Annotated[t.Optional[str], Meta(description="Profile picture url")] = None
    gender: Annotated[t.Optional[str], Meta(description="Profile gender")] = None
    summary: Annotated[t.Optional[str], Meta(description="Profile summary text")] = None


class Experience(Struct, kw_only=True):
    key: Annotated[
        t.Optional[str], Meta(description="Identification key of the Experience.")
    ] = None
    company: Annotated[
        t.Optional[str], Meta(description="Company name of the Experience.")
    ] = None
    logo: Annotated[t.Optional[str], Meta(description="Logo of the Company")] = None
    title: Annotated[t.Optional[str], Meta(description="Title of the Experience.")] = (
        None
    )
    description: Annotated[
        t.Optional[str], Meta(description="Description of the Experience.")
    ] = None
    location: Annotated[
        t.Optional[Location], Meta(description="Location object of the Experience.")
    ] = None
    date_start: Annotated[
        t.Optional[str],
        Meta(description="Start date of the experience. type: ('datetime ISO 8601')"),
    ] = None
    date_end: Annotated[
        t.Optional[str],
        Meta(description="End date of the experience. type: ('datetime ISO 8601')"),
    ] = None
    skills: Annotated[
        t.Optional[list[Skill]], Meta(description="List of skills of the Experience.")
    ] = None
    certifications: t.Optional[list[GeneralEntitySchema]]
    courses: t.Optional[list[GeneralEntitySchema]]
    tasks: t.Optional[list[GeneralEntitySchema]]


class Education(Struct, kw_only=True):
    key: Annotated[
        t.Optional[str], Meta(description="Identification key of the Education.")
    ] = None
    school: Annotated[
        t.Optional[str], Meta(description="School name of the Education.")
    ] = None
    logo: Annotated[t.Optional[str], Meta(description="Logo of the School")] = None
    title: Annotated[t.Optional[str], Meta(description="Title of the Education.")] = (
        None
    )
    description: Annotated[
        t.Optional[str], Meta(description="Description of the Education.")
    ] = None
    location: Annotated[
        t.Optional[Location], Meta(description="Location object of the Education.")
    ] = None
    date_start: Annotated[
        t.Optional[str],
        Meta(description="Start date of the Education. type: ('datetime ISO 8601')"),
    ] = None
    date_end: Annotated[
        t.Optional[str],
        Meta(description="End date of the Education. type: ('datetime ISO 8601')"),
    ] = None
    skills: Annotated[
        t.Optional[list[Skill]], Meta(description="List of skills of the Education.")
    ] = None
    certifications: t.Optional[list[GeneralEntitySchema]]
    courses: t.Optional[list[GeneralEntitySchema]]
    tasks: t.Optional[list[GeneralEntitySchema]]


class HrFlowProfile(Struct, kw_only=True):
    key: Annotated[
        t.Optional[str], Meta(description="Identification key of the Profile.")
    ] = None
    reference: Annotated[
        t.Optional[str], Meta(description="Custom identifier of the Profile.")
    ] = None
    info: Annotated[
        ProfileInfo, Meta(description="Object containing the Profile's info.")
    ]
    text_language: Annotated[
        str,
        Meta(description="Code language of the Profile. type: string code ISO 639-1"),
    ]
    text: Annotated[str, Meta(description="Full text of the Profile.")]
    archived_at: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "type: datetime ISO8601, Archive date of the Profile."
                " The value is null for unarchived Profiles."
            ),
        ),
    ] = None
    updated_at: Annotated[
        t.Optional[str],
        Meta(description="type: datetime ISO8601, Last update date of the Profile."),
    ] = None
    created_at: Annotated[
        t.Optional[str],
        Meta(description="type: datetime ISO8601, Creation date of the Profile."),
    ] = None
    experiences_duration: Annotated[
        float, Meta(description="Total number of years of experience.")
    ]
    educations_duration: Annotated[
        float, Meta(description="Total number of years of education.")
    ]
    experiences: Annotated[
        t.Optional[list[Experience]],
        Meta(description="List of experiences of the Profile."),
    ] = field(default_factory=list)
    educations: Annotated[
        t.Optional[list[Education]],
        Meta(description="List of educations of the Profile."),
    ] = field(default_factory=list)
    attachments: Annotated[
        list, Meta(description="List of documents attached to the Profile.")
    ] = field(default_factory=list)
    skills: Annotated[
        t.Optional[list[Skill]], Meta(description="List of skills of the Profile.")
    ] = None
    languages: Annotated[
        t.Optional[list[GeneralEntitySchema]],
        Meta(description="List of spoken languages of the profile"),
    ] = None
    certifications: Annotated[
        t.Optional[list[GeneralEntitySchema]],
        Meta(description="List of certifications of the Profile."),
    ] = None
    courses: Annotated[
        t.Optional[list[GeneralEntitySchema]],
        Meta(description="List of courses of the Profile."),
    ] = None
    tasks: Annotated[
        t.Optional[list[GeneralEntitySchema]],
        Meta(description="List of tasks of the Profile."),
    ] = None
    interests: Annotated[
        t.Optional[list[GeneralEntitySchema]],
        Meta(description="List of interests of the Profile."),
    ] = None
    tags: Annotated[
        t.Optional[list[GeneralEntitySchema]],
        Meta(description="List of tags of the Profile."),
    ] = None
    metadatas: Annotated[
        t.Optional[list[GeneralEntitySchema]],
        Meta(description="List of metadatas of the Profile."),
    ] = None
    labels: Annotated[
        t.Optional[list[Label]], Meta(description="List of labels of the Profile.")
    ] = None


class ResumeToParse(Struct):
    raw: bytes
    content_type: str


class HrFlowProfileParsing(Struct):
    reference: Annotated[
        t.Optional[str], Meta(description="Custom identifier of the Profile.")
    ]
    created_at: Annotated[
        str, Meta(description="type: datetime ISO8601, Creation date of the Profile.")
    ]
    resume: ResumeToParse
    tags: Annotated[
        list[GeneralEntitySchema], Meta(description="List of tags of the Profile.")
    ]
    metadatas: Annotated[
        list[GeneralEntitySchema],
        Meta(description="List of metadatas of the Profile."),
    ]



import typing as t
from enum import Enum
from logging import LoggerAdapter

import requests
from msgspec import Meta, Struct
from msgspec import json as msgspec_json
from typing_extensions import Annotated

from hrflow_connectors.v2.connectors.workable.schemas import (
    WorkableCandidate,
    WorkableJob,
)
from hrflow_connectors.v2.core.common import Entity
from hrflow_connectors.v2.core.warehouse import (
    Aisle,
    Criterias,
    ReadOperation,
    WriteOperation,
    merge,
)

WORKABLE_BASE_URL = "https://{subdomain}.workable.com/spi/v3"


class AuthParameters(Struct):
    api_access_token: Annotated[
        str,
        Meta(
            description=(
                "The API access token for the Workable account, which can be generated"
                " from the Integrations section in the Workable backend settings."
            )
        ),
    ]
    subdomain: Annotated[
        str,
        Meta(
            description=(
                "The subdomain of the Workable account, can be retrieved with a GET"
                " request to the /account endpoint"
            ),
        ),
    ]


class JobState(str, Enum):
    draft = "draft"
    published = "published"
    archived = "archived"
    closed = "closed"


class JobsReadParameters(Struct, omit_defaults=True):
    state: Annotated[
        t.Optional[JobState],
        Meta(
            description=(
                "Returns jobs with the current state. Possible values (draft,"
                " published, archived & closed)."
            ),
        ),
    ] = None
    since_id: Annotated[
        t.Optional[str],
        Meta(
            description="Returns jobs with ID greater than the specified value.",
        ),
    ] = None
    max_id: Annotated[
        t.Optional[str],
        Meta(
            description="Returns jobs with ID less than the specified value.",
        ),
    ] = None
    created_after: Annotated[
        t.Optional[t.Union[str, int]],
        Meta(
            description="Returns jobs created after the specified timestamp/date time.",
        ),
    ] = None
    updated_after: Annotated[
        t.Optional[t.Union[str, int]],
        Meta(
            description="Returns jobs updated after the specified timestamp/date time.",
        ),
    ] = None


class ReadProfileParameters(Struct, omit_defaults=True):
    email: Annotated[
        t.Optional[str],
        Meta(
            description="The email of the candidate to filter by",
        ),
    ] = None
    shortcode: Annotated[
        t.Optional[str],
        Meta(
            description="The job's system generated code",
        ),
    ] = None
    stage: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "The job's stage slug, can be retrieved from the /stages endpoint"
            ),
        ),
    ] = None
    max_id: Annotated[
        t.Optional[str],
        Meta(
            description="Returns candidates with ID less than the specified value.",
        ),
    ] = None
    created_after: Annotated[
        t.Optional[t.Union[str, int]],
        Meta(
            description=(
                "Returns candidates created after the specified timestamp/date time."
            ),
        ),
    ] = None
    updated_after: Annotated[
        t.Optional[t.Union[str, int]],
        Meta(
            description=(
                "Returns candidates updated after the specified timestamp/date time."
            ),
        ),
    ] = None


class WriteProfileParameters(Struct, omit_defaults=True):
    job_shortcode: Annotated[
        t.Optional[str],
        Meta(
            description="The job's shortcode to which the candidate is applying",
        ),
    ] = None
    stage: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "The job's stage slug, can be retrieved from the /stages endpoint"
            ),
        ),
    ] = None
    add_to_talent_pool: Annotated[
        t.Optional[bool],
        Meta(
            description="Whether to add the candidate to the talent pool",
        ),
    ] = None


class UpdateProfileParameters(Struct, omit_defaults=True):
    pass


# TODO: add pagination
def read_jobs(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: JobsReadParameters,
    incremental: bool,
    incremental_token: t.Optional[str],
) -> t.Iterator[t.Dict]:
    jobs_url = WORKABLE_BASE_URL.format(subdomain=auth_parameters.subdomain) + "/jobs"

    params = msgspec_json.decode(msgspec_json.encode(parameters), type=dict)
    # params["include_fields"] = "shortcode"

    headers = {
        "Authorization": f"Bearer {auth_parameters.api_access_token}",
    }

    jobs = []
    url = jobs_url

    while True:
        response = requests.get(url, headers=headers, params=params)

        if not response.ok:
            adapter.error(f"Failed to read Workable jobs reason : {response.text}")
            raise Exception("Failed to read Workable jobs")

        jobs.extend(response.json().get("jobs", []))
        next_url = response.json().get("paging", {}).get("next")
        if not next_url:
            break
        url = next_url

    for job in jobs:
        full_job_response = requests.get(
            f"{jobs_url}/{job['shortcode']}",
            headers=headers,
        )
        if not full_job_response.ok:
            adapter.error(
                f"Failed to read Workable job reason : {full_job_response.text}"
            )

        yield full_job_response.json()


def read_profiles(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: ReadProfileParameters,
    incremental: bool,
    incremental_token: t.Optional[str],
) -> t.Iterator[t.Dict]:
    profiles_url = (
        WORKABLE_BASE_URL.format(subdomain=auth_parameters.subdomain) + "/candidates"
    )

    params = msgspec_json.decode(msgspec_json.encode(parameters), type=dict)

    headers = {
        "Authorization": f"Bearer {auth_parameters.api_access_token}",
    }

    profiles = []
    url = profiles_url

    while True:
        response = requests.get(url, headers=headers, params=params)

        if not response.ok:
            adapter.error(f"Failed to read Workable profiles reason : {response.text}")
            raise Exception("Failed to read Workable profiles")

        profiles.extend(response.json().get("candidates", []))
        next_url = response.json().get("paging", {}).get("next")
        if not next_url:
            break
        url = next_url

    for profile in profiles:
        full_profile_response = requests.get(
            f"{profiles_url}/{profile['id']}",
            headers=headers,
        )
        if not full_profile_response.ok:
            adapter.error(
                f"Failed to read Workable profile reason : {full_profile_response.text}"
            )

        full_profile = full_profile_response.json()["candidate"]

        resume_url = full_profile.get("resume_url")
        if resume_url:
            resume_response = requests.get(resume_url)
            if resume_response.ok:
                full_profile["resume"] = resume_response.content

        yield full_profile


def write_profiles(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: WriteProfileParameters,
    items: t.Iterable[t.Dict],
) -> t.List[t.Dict]:
    failed_profiles = []

    params = {}
    if parameters.stage:
        params = {"stage": parameters.stage}

    for profile in items:
        if parameters.job_shortcode:
            post_candidate_url = (
                WORKABLE_BASE_URL.format(subdomain=auth_parameters.subdomain)
                + f"/jobs/{parameters.job_shortcode}/candidates"
            )
        if parameters.add_to_talent_pool:
            post_candidate_url = (
                WORKABLE_BASE_URL.format(subdomain=auth_parameters.subdomain)
                + "/talent_pool/candidates"
            )

        response = requests.post(
            post_candidate_url,
            headers={
                "Authorization": f"Bearer {auth_parameters.api_access_token}",
            },
            params=params,
            json=profile,
        )
        if not response.ok:
            adapter.error(f"Failed to write profile reason : {response.text}")
            failed_profiles.append(profile)
    return failed_profiles


def update_profiles(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: UpdateProfileParameters,
    items: t.Iterable[t.Dict],
) -> t.List[t.Dict]:
    failed_profiles = []

    for profile in items:
        prodile_id = profile.pop("id")
        url = (
            WORKABLE_BASE_URL.format(subdomain=auth_parameters.subdomain)
            + f"/candidates/{prodile_id}"
        )

        response = requests.patch(
            url,
            headers={
                "Authorization": f"Bearer {auth_parameters.api_access_token}",
            },
            json=profile,
        )

        if not response.ok:
            adapter.error(f"Failed to update profile reason : {response.text}")
            failed_profiles.append(profile)

    return failed_profiles


JobsAisle = Aisle(
    name=Entity.job,
    schema=WorkableJob,
    read=ReadOperation(
        criterias=Criterias(
            create=JobsReadParameters,
            update=JobsReadParameters,
            archive=JobsReadParameters,
        ),
        function=merge(
            create=read_jobs,
            update=read_jobs,
            archive=read_jobs,
        ),
    ),
)

ProfilesAisle = Aisle(
    name=Entity.profile,
    schema=WorkableCandidate,
    read=ReadOperation(
        criterias=Criterias(
            create=ReadProfileParameters,
            update=ReadProfileParameters,
            archive=ReadProfileParameters,
        ),
        function=merge(
            create=read_profiles,
            update=read_profiles,
            archive=read_profiles,
        ),
    ),
    write=WriteOperation(
        criterias=Criterias(
            create=WriteProfileParameters,
            update=UpdateProfileParameters,
        ),
        function=merge(
            create=write_profiles,
            update=update_profiles,
        ),
    ),
)



from __future__ import annotations

import importlib
import inspect
import typing as t
from pathlib import Path

from hrflow_connectors.v2.core.context import MAIN_IMPORT_NAME

if t.TYPE_CHECKING:
    from hrflow_connectors.v2.core.connector import Connector  # pragma: nocover


class ConnectorImportNameNotFound(Exception):
    pass


class AmbiguousConnectorImportName(Exception):
    pass


def get_import_name(connector: Connector) -> str:
    main_module = importlib.import_module(MAIN_IMPORT_NAME.get())

    members = inspect.getmembers(main_module, lambda s: s is connector)
    if len(members) == 0:
        raise ConnectorImportNameNotFound(
            "Failed to find import name for"
            f" connector {connector.name}\nNo match found for"
            " below members"
            f" {[symbol for symbol, _ in inspect.getmembers(main_module)]}"
        )
    if len(members) > 1:
        raise AmbiguousConnectorImportName(
            "Found multiple import names for"
            f" connector {connector.name}\n"
            f" {[symbol for symbol, _ in members]}"
        )
    return members[0][0]


class NoDefFunctionNameFound(Exception):
    pass


def reindent_function_source(source: str, function_name: str):
    search_for = f"def {function_name}"
    def_line = next(
        (line for line in source.splitlines() if search_for in line),
        None,
    )
    if def_line is None:
        raise NoDefFunctionNameFound()

    start_lines_from = def_line.find(search_for)
    return "\n".join([line[start_lines_from:] for line in source.splitlines()])


KB = 1024
MAX_LOGO_SIZE_BYTES = 100 * KB
MAX_LOGO_PIXEL = 150
MIN_LOGO_PIXEL = 34
HRFLOW_CONNECTORS_RAW_GITHUB_CONTENT_BASE = (
    "https://raw.githubusercontent.com/Riminder/hrflow-connectors"
)
CONNECTORS_DIRECTORY = Path(__file__).parent.parent / "connectors"
ROOT_DIRECTORY = Path(__file__).parent.parent.parent.parent.parent


def compute_logo_path(
    name: str, subtype: str, connectors_directory: Path = CONNECTORS_DIRECTORY
) -> str:
    try:
        from PIL import Image, UnidentifiedImageError
    except ModuleNotFoundError:  # pragma: no cover
        raise Exception(
            "PIL is not found in current environment. Mind that you need to install"
            " the package with dev dependencies to use manifest utility"
        )

    connector_directory = connectors_directory / subtype
    if not connector_directory.is_dir():
        raise ValueError(
            "No directory found for connector {} in {}".format(
                name, connector_directory
            )
        )

    logo_paths = list(connector_directory.glob("logo.*"))
    if len(logo_paths) == 0:
        raise ValueError(
            "Missing logo for connector {}. Add a logo file at {} named"
            " 'logo.(png|jpeg|...)'".format(name, connector_directory)
        )
    elif len(logo_paths) > 1:
        raise ValueError(
            "Found multiple logos for connector {} => {}. Only a single one should"
            " be present".format(name, logo_paths)
        )
    logo = logo_paths[0]
    size = logo.lstat().st_size
    if size > MAX_LOGO_SIZE_BYTES:
        raise ValueError(
            "Logo size {} KB for connector {} is above maximum limit of {} KB".format(
                size // KB, name, MAX_LOGO_SIZE_BYTES // KB
            )
        )
    try:
        width, height = Image.open(logo).size
    except UnidentifiedImageError:
        raise ValueError(
            "Logo file for connector {} at {} doesn't seem to be a valid image".format(
                name, logo
            )
        )

    if width != height or width > MAX_LOGO_PIXEL or width < MIN_LOGO_PIXEL:
        raise ValueError(
            "Bad logo dimensions of ({}, {}) for connector {}. Logo should have"
            " square dimensions within range {min}x{min} {max}x{max}".format(
                width,
                height,
                name,
                min=MIN_LOGO_PIXEL,
                max=MAX_LOGO_PIXEL,
            )
        )
    return f"{HRFLOW_CONNECTORS_RAW_GITHUB_CONTENT_BASE}/master/{logo.relative_to(ROOT_DIRECTORY)}"  # noqa E501



import typing as t
from enum import Enum
from logging import LoggerAdapter

import requests
from msgspec import Meta, Struct
from msgspec.structs import asdict
from typing_extensions import Annotated

from hrflow_connectors.v2.connectors.zohorecruit.schemas import Candidate, JobOpening
from hrflow_connectors.v2.core.common import Entity, Mode
from hrflow_connectors.v2.core.warehouse import (
    Aisle,
    Criterias,
    ReadOperation,
    WriteOperation,
    merge,
)

AUTHORIZATION_URL = "{zoho_accounts_url}/oauth/v2/token"
RECORDS_URL = "{recruit_accounts_url}/recruit/v2/{module_api_name}"


DESC = (
    "Pre requisites: a Zoho Recruit account and a registered client.\nTo register go"
    " to Zoho Developer Console: https://api-console.zoho.com/ and choose the Self"
    " Client type.\nthe creation of the Self Client will provide you with a Client ID"
    " and a Client Secret.\nYou need to then click on the 'Generate Code' button and"
    " porvide the required scopes (ZohoRecruit.modules.ALL for access to all modules in"
    " Recruite) to get the Authorization Code(grant token).\nfor  more information"
    " check the following links:\n"
    "https://www.zoho.com/recruit/developer-guide/apiv2/auth-request.html#:~:text=scopes%20here.-,Self%2DClient%20option,-Use%20this%20method \n"
    "https://help.zoho.com/portal/en/kb/recruit/developer-guide/oauth-authentication/overview/articles/oauth-overview#Scopes:~:text=the%20end%20user.-,Scopes%C2%A0%C2%A0,-Zoho%20Recruit%20APIs"
)


class ZohoAccountsURL(str, Enum):
    US = "https://accounts.zoho.com"
    AU = "https://accounts.zoho.com.au"
    EU = "https://accounts.zoho.eu"
    IN = "https://accounts.zoho.in"
    CN = "https://accounts.zoho.com.cn"
    JP = "https://accounts.zoho.jp"
    CA = "https://accounts.zohocloud.ca"


class RecruitAccountsURL(str, Enum):
    US = "https://recruit.zoho.com"
    EU = "https://recruit.zoho.eu"
    IN = "https://recruit.zoho.in"
    CN = "https://recruit.zoho.com.cn"
    AU = "https://recruit.zoho.com.au"
    JP = "https://recruit.zoho.jp"
    CA = "https://recruit.zohocloud.ca"


# Mapping from ZohoAccountsURL to RecruitAccountsURL
url_mapping = {
    ZohoAccountsURL.US: RecruitAccountsURL.US,
    ZohoAccountsURL.AU: RecruitAccountsURL.AU,
    ZohoAccountsURL.EU: RecruitAccountsURL.EU,
    ZohoAccountsURL.IN: RecruitAccountsURL.IN,
    ZohoAccountsURL.CN: RecruitAccountsURL.CN,
    ZohoAccountsURL.JP: RecruitAccountsURL.JP,
    ZohoAccountsURL.CA: RecruitAccountsURL.CA,
}


class AuthParameters(Struct):
    client_id: Annotated[
        str,
        Meta(
            description="A unique ID displayed under Self Client > Client Secret.",
        ),
    ]
    client_secret: Annotated[
        str,
        Meta(
            description=(
                "A unique confidential secret displayed under Self Client > Client"
                " Secret."
            ),
        ),
    ]
    authorization_code: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "The authorization code generated during the Self Client creation, used"
                " to get the refresh token and the first access token."
            ),
        ),
    ] = None
    refresh_token: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "The refresh token is used to generate a new access token when the"
                " current access token expires."
            ),
        ),
    ] = None
    zoho_accounts_url: Annotated[
        ZohoAccountsURL,
        Meta(
            description=(
                "Zoho CRM is hosted at multiple data centers. Therefore, the API domain"
                " URL varies for each data center.\nYou must use your domain-specific"
                " Zoho Accounts URL to generate access and refresh tokens. The"
                " following are the various domains and their corresponding accounts"
                " URLs.\n US: https://accounts.zoho.com\n AU:"
                " https://accounts.zoho.com.au\n EU: https://accounts.zoho.eu\n IN:"
                " https://accounts.zoho.in\n CN: https://accounts.zoho.com.cn\n JP:"
                " https://accounts.zoho.jp\n CA: https://accounts.zohocloud.ca"
            ),
        ),
    ] = ZohoAccountsURL.EU


class SortOrder(str, Enum):
    ASC = "asc"
    DESC = "desc"


class ZohoBool(str, Enum):
    TRUE = "true"
    FALSE = "false"
    BOTH = "both"


class State(str, Enum):
    DRAFT = "draft"
    SAVE = "save"


class Type(str, Enum):
    ALL = "all"
    RECYCLE = "recycle"
    PERMANENT = "permanent"


class ReadParameters(Struct):
    fields: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "To list all the module records with respect to fields\nMultiple field"
                " API names, comma-separated.\nFor example Last_Name, Email"
            ),
        ),
    ] = None
    sort_order: Annotated[
        t.Optional[SortOrder],
        Meta(
            description=(
                "To sort the available list of records in either ascending or"
                " descending order\nasc - ascending order\ndesc - descending order"
            ),
        ),
    ] = None
    sort_by: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "To sort the available list of records based on the given field\nField"
                " API name\nExample: Email"
            ),
        ),
    ] = None
    cvid: Annotated[
        t.Optional[int],
        Meta(
            description=(
                "To get the list of records based on custom views\n{custom_view_id}"
            ),
        ),
    ] = None
    territory_id: Annotated[
        t.Optional[int],
        Meta(
            description="To get the list of records based on territory\n{territory_id}",
        ),
    ] = None
    include_child: Annotated[
        t.Optional[bool],
        Meta(
            description=(
                "To include records from the child territories.\nTrue includes child"
                " territory records.\nFalse does not include child territory"
                " records.\nThe default value is false."
            ),
        ),
    ] = None
    state: Annotated[
        t.Optional[State],
        Meta(
            description=(
                "If the value of this parameter is 'draft', then the response will only"
                " contain Draft records from the specified module. If the parameter's"
                " value is 'save', then the response will return saved records from the"
                " specified module.\n\nIf this parameter is not included in your"
                " request body, then the response will only return saved records from"
                " the specified module."
            ),
        ),
    ] = None
    converted: Annotated[
        t.Optional[ZohoBool],
        Meta(
            description=(
                "To get the list of converted records.\nThe default value is"
                " false\ntrue - get only converted records\nfalse - get only"
                " non-converted records\nboth - get all records"
            ),
        ),
    ] = ZohoBool.FALSE
    approved: Annotated[
        t.Optional[ZohoBool],
        Meta(
            description=(
                "To get the list of approved records.\nThe default value is true\ntrue"
                " - get only approved records\nfalse - get only non-approved"
                " records\nboth - get all records"
            ),
        ),
    ] = ZohoBool.TRUE


class ReadDeleteParameters(Struct):
    type: Annotated[
        t.Optional[Type],
        Meta(
            description=(
                "All\nTo get the list of all deleted records\nRecycle\nTo get the list"
                " of deleted records from recycle bin\nPermanent\nTo get the list of"
                " permanently deleted records"
            ),
        ),
    ] = Type.ALL


class WriteParameters(Struct):
    pass


class WriteDeleteParameters(Struct):
    wf_trigger: Annotated[
        t.Optional[bool],
        Meta(
            description=(
                "Represents if the workflow rules are to be triggered upon record"
                " deletion. The value true triggers workflows. The value false does not"
                " trigger workflows. The default value is true."
            ),
        ),
    ] = True


def get_refresh_token_and_first_access_token(
    adapter, authorization_url, client_id, client_secret, authorization_code
):
    request_data = {
        "client_id": client_id,
        "client_secret": client_secret,
        "grant_type": "authorization_code",
        "code": authorization_code,
    }

    response = requests.post(authorization_url, data=request_data)
    if response.status_code == 200:
        response = response.json()
        if "error" in response:
            if response["error"] == "invalid_code":
                adapter.info(
                    "The grant token has expired or  have already been used, use"
                    " the refresh token from the previous response to get a new"
                    " access token"
                )
                return None, None

            elif response["error"] == "invalid_client":
                adapter.error("The client ID or client secret is invalid")
                return None, None

    access_token = response.json()["access_token"]
    refresh_token = response.json()["refresh_token"]
    return access_token, refresh_token


def refresh_access_token(
    adapter,
    authorization_url,
    client_id,
    client_secret,
    refresh_token,
):
    request_data = {
        "client_id": client_id,
        "client_secret": client_secret,
        "grant_type": "refresh_token",
        "refresh_token": refresh_token,
    }

    response = requests.post(authorization_url, data=request_data)
    if response.status_code == 200:
        response_data = response.json()
        if "error" in response_data:
            if response_data["error"] == "invalid_code":
                adapter.error(
                    "The refresh token to generate a new access token is wrong or"
                    " revoked."
                )
            elif response_data["error"] == "invalid_client":
                adapter.error("The client ID or client secret is invalid")
        return response_data["access_token"]

    else:
        raise Exception(
            f"Failed to get access token: {response.text}, response status code"
            f" {response.status_code}"
        )


def generic_read(
    action: Mode,
    module_api_name: str,
):
    def __pull_records(
        adapter: LoggerAdapter,
        auth_parameters: AuthParameters,
        parameters: ReadParameters,
        incremental: bool,
        incremental_token: t.Optional[str],
    ) -> t.Iterable[t.Dict]:
        authorization_url = AUTHORIZATION_URL.format(
            zoho_accounts_url=auth_parameters.zoho_accounts_url
        )
        recruit_accounts_url = url_mapping[auth_parameters.zoho_accounts_url]
        url = RECORDS_URL.format(
            recruit_accounts_url=recruit_accounts_url, module_api_name=module_api_name
        )

        updated_date_field = (
            "Modified_Time" if module_api_name == "JobOpenings" else "Updated_On"
        )

        if auth_parameters.refresh_token:
            refresh_token = auth_parameters.refresh_token
            access_token = refresh_access_token(
                adapter,
                authorization_url,
                auth_parameters.client_id,
                auth_parameters.client_secret,
                refresh_token,
            )
        else:
            access_token, refresh_token = get_refresh_token_and_first_access_token(
                adapter,
                authorization_url,
                auth_parameters.client_id,
                auth_parameters.client_secret,
                auth_parameters.authorization_code,
            )

        params = asdict(parameters)
        state = params.pop("state", None)
        if state:
            params["$state"] = state

        headers = {"Authorization": f"Zoho-oauthtoken {access_token}"}
        while True:
            response = requests.get(url, headers=headers, params=params)
            if response.status_code == 200:
                response = response.json()
                data = response["data"]
                for record in data:
                    if (
                        action == Mode.create
                        and record[updated_date_field] != record["Created_Time"]
                    ) or (
                        action == Mode.update
                        and record[updated_date_field] == record["Created_Time"]
                    ):
                        continue

                    full_record_response = requests.get(
                        f"{url}/{record['id']}", headers=headers
                    )
                    if full_record_response.status_code == 200:
                        yield full_record_response.json()["data"][0]
                    else:
                        adapter.error(
                            f"Failed to fetch record: {full_record_response.text},"
                            f" response status code {full_record_response.status_code}"
                        )
                if not response["info"]["more_records"]:
                    break
                params["page"] = response["info"]["page"] + 1

            elif response.status_code == 401:
                access_token = refresh_access_token(
                    adapter,
                    authorization_url,
                    auth_parameters.client_id,
                    auth_parameters.client_secret,
                    refresh_token,
                )
                if not access_token:
                    raise Exception("Failed to refresh access token")
                headers = {"Authorization": f"Zoho-oauthtoken {access_token}"}
            else:
                adapter.error(
                    f"Failed to fetch records: {response.text}, response status code"
                    f" {response.status_code}"
                )
                break

    return __pull_records


def generic_read_deleted(
    module_api_name: str,
):
    def __pull_deleted_records(
        adapter: LoggerAdapter,
        auth_parameters: AuthParameters,
        parameters: ReadDeleteParameters,
        incremental: bool,
        incremental_token: t.Optional[str],
    ) -> t.Iterable[t.Dict]:
        authorization_url = AUTHORIZATION_URL.format(
            zoho_accounts_url=auth_parameters.zoho_accounts_url
        )
        recruit_accounts_url = url_mapping[auth_parameters.zoho_accounts_url]
        url = (
            RECORDS_URL.format(
                recruit_accounts_url=recruit_accounts_url,
                module_api_name=module_api_name,
            )
            + "/deleted"
        )

        if auth_parameters.refresh_token:
            refresh_token = auth_parameters.refresh_token
            access_token = refresh_access_token(
                adapter,
                authorization_url,
                auth_parameters.client_id,
                auth_parameters.client_secret,
                refresh_token,
            )
        else:
            access_token, refresh_token = get_refresh_token_and_first_access_token(
                adapter,
                authorization_url,
                auth_parameters.client_id,
                auth_parameters.client_secret,
                auth_parameters.authorization_code,
            )

        params = asdict(parameters)
        headers = {"Authorization": f"Zoho-oauthtoken {access_token}"}
        while True:
            response = requests.get(url, headers=headers, params=params)
            if response.status_code == 200:
                response = response.json()
                records = response["data"]
                for record in records:
                    yield record

                if not response["info"]["more_records"]:
                    break
                params["page"] = response["info"]["page"] + 1
            elif response.status_code == 204:
                adapter.info("No records found")
                break
            elif response.status_code == 401:
                access_token = refresh_access_token(
                    adapter,
                    authorization_url,
                    auth_parameters.client_id,
                    auth_parameters.client_secret,
                    refresh_token,
                )
                if not access_token:
                    raise Exception("Failed to refresh access token")
                headers = {"Authorization": f"Zoho-oauthtoken {access_token}"}
            else:
                adapter.error(
                    f"Failed to fetch records: {response.text}, response status code"
                    f" {response.status_code}"
                )
                break

    return __pull_deleted_records


def generic_write(
    module_api_name: str,
):
    def __push_records(
        adapter: LoggerAdapter,
        auth_parameters: AuthParameters,
        parameters: WriteParameters,
        items: t.Iterable[t.Dict],
    ) -> t.List[t.Dict]:
        failed_records = []
        authorization_url = AUTHORIZATION_URL.format(
            zoho_accounts_url=auth_parameters.zoho_accounts_url
        )
        recruit_accounts_url = url_mapping[auth_parameters.zoho_accounts_url]
        url = RECORDS_URL.format(
            recruit_accounts_url=recruit_accounts_url, module_api_name=module_api_name
        )

        if auth_parameters.refresh_token:
            refresh_token = auth_parameters.refresh_token
            access_token = refresh_access_token(
                adapter,
                authorization_url,
                auth_parameters.client_id,
                auth_parameters.client_secret,
                refresh_token,
            )
        else:
            access_token, refresh_token = get_refresh_token_and_first_access_token(
                adapter,
                authorization_url,
                auth_parameters.client_id,
                auth_parameters.client_secret,
                auth_parameters.authorization_code,
            )

        headers = {"Authorization": f"Zoho-oauthtoken {access_token}"}

        for record in items:
            _input = {"data": [record]}
            response = requests.post(url, headers=headers, json=_input)
            if response.status_code == 201:
                id = response.json()["data"][0]["details"]["id"]
                adapter.info(f"Record {id} created")
            elif response.status_code == 401:
                access_token = refresh_access_token(
                    adapter,
                    authorization_url,
                    auth_parameters.client_id,
                    auth_parameters.client_secret,
                    refresh_token,
                )
                if not access_token:
                    raise Exception("Failed to refresh access token")
                headers = {"Authorization": f"Zoho-oauthtoken {access_token}"}
                response = requests.post(url, headers=headers, json=_input)
                if response.status_code == 200:
                    adapter.info(f"Record {id} created")
                else:
                    adapter.error(
                        f"Failed to create record: {response.text}, response status"
                        f" code {response.status_code}"
                    )
                    failed_records.append(record)
            else:
                adapter.error(
                    f"Failed to create record: {response.text}, response status code"
                    f" {response.status_code}"
                )
                failed_records.append(record)
        return failed_records

    return __push_records


def generic_update(
    module_api_name: str,
):
    def __update_records(
        adapter: LoggerAdapter,
        auth_parameters: AuthParameters,
        parameters: WriteParameters,
        items: t.Iterable[t.Dict],
    ) -> t.List[t.Dict]:
        failed_records = []
        authorization_url = AUTHORIZATION_URL.format(
            zoho_accounts_url=auth_parameters.zoho_accounts_url
        )
        recruit_accounts_url = url_mapping[auth_parameters.zoho_accounts_url]
        url = RECORDS_URL.format(
            recruit_accounts_url=recruit_accounts_url, module_api_name=module_api_name
        )

        if auth_parameters.refresh_token:
            refresh_token = auth_parameters.refresh_token
            access_token = refresh_access_token(
                adapter,
                authorization_url,
                auth_parameters.client_id,
                auth_parameters.client_secret,
                refresh_token,
            )
        else:
            access_token, refresh_token = get_refresh_token_and_first_access_token(
                adapter,
                authorization_url,
                auth_parameters.client_id,
                auth_parameters.client_secret,
                auth_parameters.authorization_code,
            )

        headers = {"Authorization": f"Zoho-oauthtoken {access_token}"}
        for record in items:
            _input = {"data": [record]}
            response = requests.put(url, headers=headers, json=_input)
            if response.status_code == 200:
                id = response.json()["data"][0]["details"]["id"]
                adapter.info(f"Record {id} updated")
            elif response.status_code == 401:
                access_token = refresh_access_token(
                    adapter,
                    authorization_url,
                    auth_parameters.client_id,
                    auth_parameters.client_secret,
                    refresh_token,
                )
                if not access_token:
                    raise Exception("Failed to refresh access token")
                headers = {"Authorization": f"Zoho-oauthtoken {access_token}"}
                response = requests.put(url, headers=headers, json=_input)
                if response.status_code == 200:
                    adapter.info(f"Record {id} updated")
                else:
                    adapter.error(
                        f"Failed to update record: {response.text}, response status"
                        f" code {response.status_code}"
                    )
                    failed_records.append(record)
            else:
                adapter.error(
                    f"Failed to update record: {response.text}, response status code"
                    f" {response.status_code}"
                )
                failed_records.append(record)
        return failed_records

    return __update_records


def generic_delete(
    module_api_name: str,
):
    def __delete_records(
        adapter: LoggerAdapter,
        auth_parameters: AuthParameters,
        parameters: WriteDeleteParameters,
        items: t.Iterable[t.Dict],
    ) -> t.List[t.Dict]:
        failed_records = []
        authorization_url = AUTHORIZATION_URL.format(
            zoho_accounts_url=auth_parameters.zoho_accounts_url
        )
        recruit_accounts_url = url_mapping[auth_parameters.zoho_accounts_url]
        url = RECORDS_URL.format(
            recruit_accounts_url=recruit_accounts_url, module_api_name=module_api_name
        )

        if auth_parameters.refresh_token:
            refresh_token = auth_parameters.refresh_token
            access_token = refresh_access_token(
                adapter,
                authorization_url,
                auth_parameters.client_id,
                auth_parameters.client_secret,
                refresh_token,
            )
        else:
            access_token, refresh_token = get_refresh_token_and_first_access_token(
                adapter,
                authorization_url,
                auth_parameters.client_id,
                auth_parameters.client_secret,
                auth_parameters.authorization_code,
            )

        headers = {"Authorization": f"Zoho-oauthtoken {access_token}"}
        params = asdict(parameters)
        for record in items:
            response = requests.delete(
                f"{url}?ids={record['id']}", headers=headers, params=params
            )
            if response.status_code == 200:
                id = response.json()["data"][0]["details"]["id"]
                adapter.info(f"Record {id} deleted")
            elif response.status_code == 401:
                access_token = refresh_access_token(
                    adapter,
                    authorization_url,
                    auth_parameters.client_id,
                    auth_parameters.client_secret,
                    refresh_token,
                )
                if not access_token:
                    raise Exception("Failed to refresh access token")
                headers = {"Authorization": f"Zoho-oauthtoken {access_token}"}
                response = requests.delete(f"{url}?ids={record['id']}", headers=headers)
                if response.status_code == 200:
                    adapter.info(f"Record {id} deleted")
                else:
                    adapter.error(
                        f"Failed to delete record: {response.text}, response status"
                        f" code {response.status_code}"
                    )
                    failed_records.append(record)
            else:
                adapter.error(
                    f"Failed to delete record: {response.text}, response status code"
                    f" {response.status_code}"
                )
                failed_records.append(record)
        return failed_records

    return __delete_records


JobsAisle = Aisle(
    name=Entity.job,
    schema=JobOpening,
    read=ReadOperation(
        criterias=Criterias(
            create=ReadParameters,
            archive=ReadDeleteParameters,
            update=ReadParameters,
        ),
        function=merge(
            create=generic_read(module_api_name="JobOpenings", action=Mode.create),
            archive=generic_read_deleted("JobOpenings"),
            update=generic_read(module_api_name="JobOpenings", action=Mode.update),
        ),
    ),
)

ProfilesAisle = Aisle(
    name=Entity.profile,
    schema=Candidate,
    read=ReadOperation(
        criterias=Criterias(
            create=ReadParameters,
            archive=ReadDeleteParameters,
            update=ReadParameters,
        ),
        function=merge(
            create=generic_read(module_api_name="Candidates", action=Mode.create),
            archive=generic_read_deleted("Candidates"),
            update=generic_read(module_api_name="Candidates", action=Mode.update),
        ),
    ),
    write=WriteOperation(
        criterias=Criterias(
            create=WriteParameters,
            update=WriteParameters,
            archive=WriteDeleteParameters,
        ),
        function=merge(
            create=generic_write("Candidates"),
            update=generic_update("Candidates"),
            archive=generic_delete("Candidates"),
        ),
    ),
)



import typing as t

from hrflow_connectors.v2.connectors.teamtailor.warehouse import TeamTailorWarehouse
from hrflow_connectors.v2.core.common import Direction, Entity, Mode
from hrflow_connectors.v2.core.connector import Connector, ConnectorType, Flow


def format_job(teamtailor_job: t.Dict) -> t.Dict:
    """
    Format a Teamtailor job object into a Hrflow job object
    Args:
        teamtailor_job: Teamtailor job object from the list of jobs pulled
    Returns:
        HrflowJob: a job in the HrFlow job object form
    """
    job_attributes = teamtailor_job["attributes"]
    t = lambda name, value: dict(name=name, value=value)
    job = dict(
        reference=teamtailor_job.get("id"),
        name=job_attributes.get("title"),
        summary=job_attributes.get("body"),
        created_at=job_attributes.get("created-at"),
        updated_at=job_attributes.get("updated-at"),
        url=teamtailor_job.get("links", {}).get("careersite-job-url"),
        location=teamtailor_job.get("location"),
        sections=[
            dict(
                name="job_description",
                title="Job Description",
                description=job_attributes.get("body"),
            ),
            dict(
                name="job_pitch",
                title="Job Pitch",
                description=job_attributes.get("pitch"),
            ),
        ],
        tags=[
            t("start-date", job_attributes.get("start-date")),
            t("end-date", job_attributes.get("end-date")),
            t("status", job_attributes.get("status")),
            t("human-status", job_attributes.get("human-status")),
            t("language-code", job_attributes.get("language-code")),
            t("employment-type", job_attributes.get("employment-type")),
            t("employment-level", job_attributes.get("employment-level")),
            t("remote-status", job_attributes.get("remote-status")),
            t("salary-time-unit", job_attributes.get("salary-time-unit")),
            t("min-salary", job_attributes.get("min-salary")),
            t("max-salary", job_attributes.get("max-salary")),
            t("currency", job_attributes.get("currency")),
            t("internal", job_attributes.get("internal")),
            t("internal-name", job_attributes.get("internal-name")),
            t("pinned", job_attributes.get("pinned")),
            t("teamtailor_tags", job_attributes.get("tags")),
            t(
                "careersite-job-apply-url",
                teamtailor_job.get("links", {}).get("careersite-job-apply-url"),
            ),
        ],
        ranges_float=[
            dict(
                name="Salary",
                value_min=job_attributes.get("min-salary"),
                value_max=job_attributes.get("max-salary"),
                unit=job_attributes.get("currency"),
            ),
        ],
        ranges_date=[
            dict(
                name="Period",
                value_min=job_attributes.get("start-date"),
                value_max=job_attributes.get("end-date"),
            ),
        ],
    )

    return job


def format_teamtailor_item_for_archive(item: t.Dict) -> t.Dict:
    """
    Format a Teamtailor object to be archived in Hrflow
    Args:
        item (Dict): Teamtailor object (profile/job)
    Returns:
        Dict: reference to an Hrflow object to be archived
    """
    return {"reference": item.get("id")}


def format_teamtailor_profile(teamtailor_profile: t.Dict) -> t.Dict:
    """
    Format a Teamtailor profile object into a Hrflow profile object
    Args:
        teamtailor_profile: Teamtailor Profile to format
    Returns:
        Dict[str, Any]: a Hrflow formatted profile object
    """
    profile_attributes = teamtailor_profile["attributes"]
    t = lambda name, value: dict(name=name, value=value)
    profile = dict(
        reference=teamtailor_profile.get("id"),
        info=dict(
            first_name=profile_attributes.get("first-name"),
            last_name=profile_attributes.get("last-name"),
            full_name=(
                f"{profile_attributes.get('first-name')} "
                f"{profile_attributes.get('last-name')}"
            ),
            email=profile_attributes.get("email"),
            phone=profile_attributes.get("phone"),
            location=dict(text="", lat=None, lng=None),
            picture=profile_attributes.get("picture"),
            urls=[dict(type="linkedin", url=profile_attributes.get("linkedin-url"))],
            summary=profile_attributes.get("pitch"),
        ),
        updated_at=profile_attributes.get("updated-at"),
        created_at=profile_attributes.get("created-at"),
        experiences=[],
        educations=[],
        resume=teamtailor_profile.get("resume"),
        tags=[
            t("sourced", profile_attributes.get("sourced")),
            t("connected", profile_attributes.get("connected")),
            t("internal", profile_attributes.get("internal")),
            t("referred", profile_attributes.get("referred")),
            t("referring_url", profile_attributes.get("referring-url")),
            t("referring_site", profile_attributes.get("referring-site")),
            t("unsubscribed", profile_attributes.get("unsubscribed")),
            t("teamtailor_tags", profile_attributes.get("tags")),
        ],
    )
    return profile


def format_hrflow_profile(hrflow_profile: t.Dict) -> t.Dict:
    """
    Format a Hrflow profile object into a Teamtailor profile object
    Args:
        hrflow_profile: Hrflow Profile to format
    Returns:
        Dict[str, Any]: a Teamtailor formatted profile object
    """
    profile_info = hrflow_profile["info"]
    profile = {
        "first-name": profile_info.get("first_name"),
        "last-name": profile_info.get("last_name"),
        "email": profile_info.get("email"),
        "phone": profile_info.get("phone"),
        "picture": profile_info.get("picture"),
        "linkedin-url": next(
            (
                url["url"]
                for url in profile_info.get("urls", [])
                if url["type"] == "linkedin"
            ),
            None,
        ),
        "pitch": profile_info.get("summary"),
        "resume": (
            hrflow_profile["attachments"][0].get("public_url")
            if hrflow_profile["attachments"]
            else None
        ),
        "sourced": True,
    }
    return profile


def format_hrflow_profile_for_update(hrflow_profile: t.Dict) -> t.Dict:
    """
    Format a Hrflow profile object into a Teamtailor profile object for update
    Args:
        hrflow_profile: Hrflow Profile to format
    Returns:
        Dict[str, Any]: a Teamtailor formatted profile object
    """
    profile = format_hrflow_profile(hrflow_profile)
    profile["id"] = hrflow_profile["reference"]
    return profile


DESCRIPTION = (
    "The new way to attract, nurture and hire top talent. Grow faster by focusing on"
    " what matters the most — your candidates"
)

Teamtailor = Connector(
    name="Teamtailor",
    type=ConnectorType.ATS,
    subtype="teamtailor",
    description=DESCRIPTION,
    url="https://www.teamtailor.com/",
    warehouse=TeamTailorWarehouse,
    flows=(
        Flow(Mode.create, Entity.job, Direction.inbound, format=format_job),
        Flow(Mode.update, Entity.job, Direction.inbound, format=format_job),
        Flow(
            Mode.archive,
            Entity.job,
            Direction.inbound,
            format=format_teamtailor_item_for_archive,
        ),
        Flow(
            Mode.create,
            Entity.profile,
            Direction.inbound,
            format=format_teamtailor_profile,
        ),
        Flow(
            Mode.update,
            Entity.profile,
            Direction.inbound,
            format=format_teamtailor_profile,
        ),
        Flow(
            Mode.archive,
            Entity.profile,
            Direction.inbound,
            format=format_teamtailor_item_for_archive,
        ),
        Flow(
            Mode.create,
            Entity.profile,
            Direction.outbound,
            format=format_hrflow_profile,
        ),
        Flow(
            Mode.update,
            Entity.profile,
            Direction.outbound,
            format=format_hrflow_profile_for_update,
        ),
    ),
)
